#!/usr/bin/env python

"""Produce CSV listings of Ethernet connected chip positions."""

import argparse

from spinner.diagrams.machine_map import draw_machine_map, \
	get_machine_map_aspect_ratio

from spinner.utils import folded_torus

from spinner import transforms
from spinner import topology

from spinner.scripts import arguments
from spinner.scripts.contexts import PDFContextManager, PNGContextManager


def main(args=None):
	parser = argparse.ArgumentParser(
		description="Produce CSV listings of Ethernet connected chip physical and "
		            "network positions.")
	arguments.add_version_args(parser)
	arguments.add_topology_args(parser)
	arguments.add_cabinet_args(parser)
	
	# Process command-line arguments
	args = parser.parse_args(args)
	(w, h), transformation, uncrinkle_direction, folds =\
		arguments.get_topology_from_args(parser, args)
	
	cabinet, num_frames = arguments.get_cabinets_from_args(parser, args)
	
	# Generate folded system
	hex_boards, folded_boards = folded_torus(w, h,
	                                         transformation,
	                                         uncrinkle_direction,
	                                         folds)
	
	# Divide into cabinets
	cabinetised_boards = transforms.cabinetise(folded_boards,
	                                           cabinet.num_cabinets,
	                                           num_frames,
	                                           cabinet.boards_per_frame)
	cabinetised_boards = transforms.remove_gaps(cabinetised_boards)
	
	# Generate the output
	print("cabinet,frame,board,x,y")
	b2c = dict(cabinetised_boards)
	for board, hex_coord in sorted(hex_boards, key=(lambda v: topology.to_xy(v[1]))):
		x, y = topology.to_xy(topology.board_to_chip(hex_coord))
		c, f, b = b2c[board]
		print(",".join(map(str, [c,f,b, x,y])))
	
	
	return 0


if __name__=="__main__":  # pragma: no cover
	import sys
	sys.exit(main())





"""An interactive, graphical tool which can guide a user through a predefined
list of wiring instructions."""

import colorsys

import subprocess

import traceback

from itertools import cycle

from six import iteritems, itervalues, next

import cairocffi as cairo

from math import pi

try:  # pragma: no cover
	# Python 3
	from tkinter import Tk, Label
except ImportError:  # pragma: no cover
	# Python 2
	from Tkinter import Tk, Label

from PIL import Image, ImageTk

from spinner.topology import Direction

from spinner.diagrams.machine import MachineDiagram


class InteractiveWiringGuide(object):
	"""
	An interactive, graphical tool which can guide a user through a predefined
	list of wiring instructions.
	
	After initialisation, calling "main()" will block while the GUI runs.
	
	Features:
	* Cycle through a list of wiring instructions
	* Display a full view of the system being wired up
	* Display close-up views of pairs of boards being connected
	* Illuminate an LED on boards to be connected
	* Read instructions using text-to-speech
	* Colour code diagrams by wire-length
	"""
	
	# Colour of highlights for top-left and bottom-right ends of the current
	# cable.
	TOP_LEFT_COLOUR     = (1.0, 0.0, 0.0, 1.0)
	BOTTOM_RIGHT_COLOUR = (0.0, 0.0, 1.0, 1.0)
	
	# Width of the zoomed-in areas as a fraction of the display width
	ZOOMED_VIEW_WIDTH = 0.25
	
	# Height of each row of text under the drawings.
	TEXT_ROW_HEIGHT = 0.07
	
	# Zoom-out from zoomed in areas by this ratio
	ZOOMED_MARGINS = 0.8
	
	# Poll interval in ms between checking the if the current wire has been
	# inserted or not.
	POLL_INTERVAL_MS = 500
	
	def __init__( self
	            , cabinet
	            , wire_lengths
	            , wires
	            , starting_wire=0
	            , focus=[]
	            , bmp_controller=None
	            , bmp_led=7
	            , wiring_probe=None
	            , auto_advance=True
	            , use_tts=True
	            , show_installed_wires=True
	            , show_future_wires=False
	            , timing_logger=None
	            ):
		"""
		cabinet defines the size of cabinets in the system.
		
		wire_lengths is a list of all valid wire lengths.
		
		wires is a list [(src, dst, length), ...] where src and dst are tuples
		(cabinet, rack, slot, socket) and length is a length included in
		wire_lengths or is None indicating that the wire should be disconnected.
		
		starting_wire is the index of the first wire to be inserted. This could be
		used, e.g. to resume installation at a specified point.
		
		focus is a set of arguments (cabinet, frame, board) to use to select the
		area of interest for the central system diagram.
		
		bmp_ips is a dictionary {board_position: ip} where board_position is either
		a tuple (cabinet, rack, slot) or (cabinet, rack) where the former will be
		used if both are available. The IP should be given as a string.
		
		bmp_led specifies which LED will be illuminated for boards where an IP is
		known.
		
		wiring_probe is an optional WiringProbe object which will be used to
		auto-advance through the wiring when wires are inserted correctly.
		
		auto_advance specifies whether auto-advance will be enabled initially.
		
		use_tts specifies whether text-to-spech will be used to announce
		instructions.
		
		show_installed_wires selects whether already-installed wires should be shown
		(feintly) at all times.
		
		show_future_wires selects whether to-be-installed wires should be shown
		(feintly) at all times.
		
		timing_logger is a TimingLogger into which the cabling connections made
		will be logged. If None, no timings are logged.
		"""
		
		self.cabinet      = cabinet
		self.wire_lengths = wire_lengths
		self.wires        = wires
		
		self.cur_wire  = starting_wire
		assert 0 <= self.cur_wire < len(self.wires), "Starting wire out of range."
		
		self.focus = focus
		
		self.bmp_controller = bmp_controller
		
		self.bmp_led = bmp_led
		
		self.wiring_probe = wiring_probe
		
		self.auto_advance = auto_advance
		
		self.use_tts = use_tts
		
		self.show_installed_wires = show_installed_wires
		self.show_future_wires    = show_future_wires
		
		self.timing_logger = timing_logger
		
		# Human readable names for each socket
		self.socket_names = {d: d.name.replace("_", " ") for d in Direction}
		
		# An infinately cycling iterator over all the boards in the machine.
		self.board_iter = iter(cycle(set((c, f, b)
		                                 for ((c, f, b, _1), _2, _3)
		                                 in self.wires)))
		
		# A reference to any running TTS job
		self.tts_process = None
		
		# A dict {fn: ([key, ...], description), ...} giving the keybindings and
		# homan-readable help string for all bound keyboard shortcuts.
		self.bindings = {}
		
		# Set up the Tk UI
		self.tk = Tk()
		self._init_ui()
		
		# Get started
		if self.timing_logger is not None:
			self.timing_logger.logging_started()
		self.go_to_wire(starting_wire)
	
	
	def _init_ui(self):
		"""Initialise the Tk interface."""
		self.tk.wm_title("SpiNNer Interactive Wiring Guide")
		self.tk.geometry("1024x768")
		
		# Add a label widget into which the rendered UI is drawn
		self.widget = Label()
		self.widget.pack(expand=True, fill="both")
		
		# Used to avoid unecessary redraws on window resizes where nothing actually
		# changes.
		self._old_size = (None, None)
		
		# A flag which indicates whether, on the last poll of cable insertion, the
		# cable was found to be inserted but inserted into the wrong socket.
		self.connected_incorrectly = False
		
		# Set up a timer to poll for cable insertion
		self.tk.after(InteractiveWiringGuide.POLL_INTERVAL_MS, self._poll_wiring_probe)
		
		# Handle window events
		self.tk.bind("<Configure>", self._on_resize)  # Resize
		self.tk.protocol("WM_DELETE_WINDOW", self._on_close)  # Window closed
		
		# Setup key bindings
		self.bindings = {
			self._on_next:
				(["<Button-1>", "<Key-space>", "<Key-Down>", "<Key-Right>",
				  "<Key-Return>", "<Key-Tab>"],
				 "Next wire"),
			self._on_prev:
				(["<Button-3>", "<Key-Up>", "<Key-Left>", "<Key-BackSpace>"],
				 "Previous wire"),
			self._on_skip_next: (["<Key-Next>"], "Skip forward"),
			self._on_skip_prev: (["<Key-Prior>"], "Skip backward"),
			self._on_first: (["<Key-Home>"], "First wire"),
			self._on_last: (["<Key-End>"], "Last wire"),
			self._on_tts_toggle: (["<Key-t>"], "Toggle speech"),
		}
		
		if self.wiring_probe is not None:  # pragma: no branch
			self.bindings[self._on_auto_advance_toggle] = (
				["<Key-a>"], "Toggle auto-advance")
		
		if self.timing_logger is not None: # pragma: no branch
			self.bindings[self._on_pause] = (["<Key-p>"], "Pause")
		
		for fn, (keys, help) in iteritems(self.bindings):
			for key in keys:
				self.tk.bind(key, fn)
	
	
	def go_to_wire(self, wire):
		"""
		Advance to a specific wire.
		"""
		last_wire = self.cur_wire
		self.cur_wire = wire
		
		# Reset the incorrect connection flag
		self.connected_incorrectly = False
		
		# Update LEDs
		self.set_leds(last_wire, False)
		self.set_leds(self.cur_wire, True)
		
		# Log the start of *insertion* of a new wire
		src, dst, length = self.wires[wire]
		if self.timing_logger is not None:
			self.timing_logger.unpause()
			if length is not None:
				sc, sf, sb, sd = self.wires[wire][0]
				dc, df, db, dd = self.wires[wire][1]
				self.timing_logger.connection_started(sc, sf, sb, sd,
				                                      dc, df, db, dd)
		
		# Announce via TTS the distance relative to the last position
		if self.use_tts:
			self.tts_delta(last_wire, self.cur_wire)
	
	
	def set_leds(self, wire, state):
		"""
		Set the LEDs for the given wire index to the given state (assuming the
		board's IP is known).
		"""
		try:
			if self.bmp_controller is not None:
				for c,f,b,p in self.wires[wire][:2]:
					self.bmp_controller.set_led(self.bmp_led, state, c, f, b)
		except:
			# Quit if this goes wrong
			self.tk.destroy()
			raise
	
	
	def tts_delta(self, last_wire, this_wire):
		"""
		Announce via TTS a brief instruction indicating what the next wire should be
		in terms of the difference to the previous wire.
		
		Changes are announced relative to the last wire.
		"""
		message = ""
		
		# Announce wire-length changes
		last_length = self.wires[last_wire][2]
		this_length = self.wires[this_wire][2]
		
		if last_length != this_length:
			if this_length is None:
				message += "Disconnect cable. "
			else:
				message += "%s meter cable. "%(("%0.2f"%this_length).rstrip(".0"))
		
		# Announce which ports are being connected
		this_tl = self._top_left_socket(this_wire)
		this_br = self._bottom_right_socket(this_wire)
		
		message += self.socket_names[this_tl[3]]
		message += " going "
		message += self.socket_names[this_br[3]]
		message += "."
		
		self._tts_speak(message)
	
	
	def _tts_speak(self, text, wpm = 250):
		"""
		Speak the supplied string, interrupting whatever was already being said.
		Non-blocking.
		"""
		# Kill previous instances
		if self.tts_process is not None and self.tts_process.poll() is None:
			self.tts_process.terminate()
		
		# Speak the required text.
		self.tts_process = subprocess.Popen( ["espeak", "-s", str(wpm), text]
		                                   , stdout = subprocess.PIPE
		                                   , stderr = subprocess.PIPE
		                                   )
	
	
	def _get_wire_colour(self, length):
		"""
		Get the RGB colour (as a tuple) for wires of the specified length.
		
		Colours are allocated evenly across the spectrum. Wires to be removed are
		shown in black.
		"""
		if length is None:
			return (0.0, 0.0, 0.0)
		
		index = sorted(self.wire_lengths).index(length)
		
		hue = index / float(len(self.wire_lengths))
		
		return colorsys.hsv_to_rgb(hue, 1.0, 0.5)
	
	
	def _top_left_socket(self, wire):
		"""
		Return the (c,r,s,d) for the top-left socket for the current wire.
		"""
		
		src, dst, length = self.wires[wire]
		
		return min([src, dst], key=(lambda v: (-v[0],  # Right-to-left
		                                       +v[1],  # Top-to-bottom
		                                       -v[2])))  # Right-to-left
	
	
	def _bottom_right_socket(self, wire):
		"""
		Return the (c,r,s,d) for the bottom-right socket for the current wire.
		"""
		
		src, dst, length = self.wires[wire]
		
		return max([src, dst], key=(lambda v: (-v[0],  # Right-to-left
		                                       +v[1],  # Top-to-bottom
		                                       -v[2])))  # Right-to-left
	
	
	def _get_machine_diagram(self):
		"""
		Get the MachineDiagram ready to draw the system's current state.
		"""
		md = MachineDiagram(self.cabinet)
		
		bg_wire = self.cabinet.board_dimensions.x / 5.0
		fg_wire = self.cabinet.board_dimensions.x / 3.0
		
		board_hl = self.cabinet.board_dimensions.x / 3.0
		wire_hl = self.cabinet.board_dimensions.x / 2.0
		
		# Wires already installed
		if self.show_installed_wires:
			for src, dst, length in self.wires[:self.cur_wire]:
				r,g,b = self._get_wire_colour(length)
				md.add_wire(src, dst, rgba = (r,g,b,0.5), width = bg_wire)
		
		# Wires still to be installed
		if self.show_future_wires:
			for src, dst, length in self.wires[self.cur_wire+1:]:
				r,g,b = self._get_wire_colour(length)
				md.add_wire(src, dst, rgba = (r,g,b,0.5), width = bg_wire)
		
		# Current wire (with a white outline)
		src, dst, length = self.wires[self.cur_wire]
		r,g,b = self._get_wire_colour(length)
		md.add_wire(src, dst, rgba = (1.0,1.0,1.0,1.0), width = fg_wire * 2)
		md.add_wire(src, dst, rgba = (r,g,b,1.0),       width = fg_wire)
		
		# Highlight source and destination
		c,r,s,d = self._top_left_socket(self.cur_wire)
		md.add_highlight(c,r,s,d, rgba = self.TOP_LEFT_COLOUR, width=wire_hl)
		md.add_highlight(c,r,s,   rgba = self.TOP_LEFT_COLOUR, width=board_hl)
		c,r,s,d = self._bottom_right_socket(self.cur_wire)
		md.add_highlight(c,r,s,d, rgba = self.BOTTOM_RIGHT_COLOUR, width=wire_hl)
		md.add_highlight(c,r,s,   rgba = self.BOTTOM_RIGHT_COLOUR, width=board_hl)
		
		return md
	
	
	def _draw_text(self, ctx, text, size, rgba = (0.0,0.0,0.0, 1.0)):
		"""
		Draw the desired text centered below (0,0).
		"""
		ctx.save()
		
		ctx.select_font_face("Sans")
		ctx.set_source_rgba(*rgba)
		ctx.set_font_size(size*0.8)
		x,y, w,h, _w,_h = ctx.text_extents(text)
		ctx.move_to(-x - w/2, -y + size*0.1)
		ctx.show_text(text)
		
		ctx.restore()
	
	
	def _draw_help_text(self, ctx, width, height, rgba = (0.5,0.5,0.5, 1.0)):
		"""
		Draw the help text along the bottom of the screen, return the height of the
		text in pixels.
		"""
		# Generate help string
		help_text = "  |  ".join("{} {}".format(keys[0], help)
		                         for keys, help in 
		                         sorted(itervalues(self.bindings),
		                                key=(lambda kh: kh[0])))
		
		ctx.save()
		
		# Determine font size which will fill the width of the screen
		ctx.select_font_face("Sans")
		ctx.set_source_rgba(*rgba)
		ctx.set_font_size(1.0)
		x,y, w,h, _w,_h = ctx.text_extents(help_text)
		scale_factor = (width * 0.95) / w
		
		# Draw the text along the bottom of the screen
		ctx.set_font_size(scale_factor)
		x,y, w,h, _w,_h = ctx.text_extents(help_text)
		ctx.move_to(x + ((width - w) / 2), height + y)
		ctx.show_text(help_text)
		
		ctx.restore()
		
		return h
	
	
	def _render_gui(self, ctx, width, height):
		"""
		Re-draw the whole GUI into the supplied Cairo context.
		"""
		# Clear the buffer background
		ctx.set_source_rgba(1.0,1.0,1.0,1.0);
		ctx.rectangle(0,0, width, height)
		ctx.fill()
		
		# Draw help text along bottom of screen.
		height -= self._draw_help_text(ctx, width, height)
		
		md = self._get_machine_diagram()
		
		# Draw the main overview image
		ctx.save()
		ctx.translate(width*self.ZOOMED_VIEW_WIDTH, 0.0)
		ctx.rectangle(0,0, width * (1.0 - (2*self.ZOOMED_VIEW_WIDTH)), height * (1 - (2*self.TEXT_ROW_HEIGHT)))
		ctx.clip()
		md.draw( ctx
		       , width * (1.0 - (2*self.ZOOMED_VIEW_WIDTH))
		       , height * (1 - (2*self.TEXT_ROW_HEIGHT))
		       , *((list(self.focus) + [None]*3)[:3] * 2)
		       )
		ctx.restore()
		
		# Draw the left zoomed-in image
		ctx.save()
		ctx.rectangle(0,0, width*self.ZOOMED_VIEW_WIDTH, height*(1-(2*self.TEXT_ROW_HEIGHT)))
		ctx.clip()
		ctx.translate( width*self.ZOOMED_VIEW_WIDTH*(1-self.ZOOMED_MARGINS)/2
		             , height*(1-(2*self.TEXT_ROW_HEIGHT))*(1-self.ZOOMED_MARGINS)/2
		             )
		ctx.scale(self.ZOOMED_MARGINS, self.ZOOMED_MARGINS)
		md.draw( ctx
		       , width*self.ZOOMED_VIEW_WIDTH
		       , height*(1 - (2*self.TEXT_ROW_HEIGHT))
		       , *(list(self._top_left_socket(self.cur_wire)[:3]) +
		           list(self.focus))
		       )
		ctx.restore()
		
		# Draw the right zoomed-in image
		ctx.save()
		ctx.translate(width*(1-self.ZOOMED_VIEW_WIDTH), 0.0)
		ctx.rectangle(0,0, width*self.ZOOMED_VIEW_WIDTH, height*(1-(2*self.TEXT_ROW_HEIGHT)))
		ctx.clip()
		ctx.translate( width*self.ZOOMED_VIEW_WIDTH*(1-self.ZOOMED_MARGINS)/2
		             , height*(1-(2*self.TEXT_ROW_HEIGHT))*(1-self.ZOOMED_MARGINS)/2
		             )
		ctx.scale(self.ZOOMED_MARGINS, self.ZOOMED_MARGINS)
		md.draw( ctx
		       , width*self.ZOOMED_VIEW_WIDTH
		       , height*(1 - (2*self.TEXT_ROW_HEIGHT))
		       , *(list(self._bottom_right_socket(self.cur_wire)[:3]) +
		           list(self.focus))
		       )
		ctx.restore()
		
		# Draw the wire length
		ctx.save()
		ctx.translate(width/2, height*(1 - (2*self.TEXT_ROW_HEIGHT)))
		length = self.wires[self.cur_wire][2]
		self._draw_text( ctx
		               , "%0.2f m"%(length)
		                 if length is not None
		                 else "Disconnect Wire"
		               , height*self.TEXT_ROW_HEIGHT
		               , rgba = self._get_wire_colour(length)
		               )
		
		# Draw the progress
		ctx.translate(0, height*self.TEXT_ROW_HEIGHT)
		self._draw_text( ctx
		               , "%d of %d (%0.1f%%)"%( self.cur_wire + 1
		                                      , len(self.wires)
		                                      , 100.0*((self.cur_wire+1)/float(len(self.wires)))
		                                      )
		               , height*self.TEXT_ROW_HEIGHT
		               )
		ctx.restore()
		
		# Draw the endpoint specifications
		for x_offset, (c,r,s,d) in [ (width * (self.ZOOMED_VIEW_WIDTH/2),     self._top_left_socket(self.cur_wire))
		                           , (width * (1-(self.ZOOMED_VIEW_WIDTH/2)), self._bottom_right_socket(self.cur_wire))
		                           ]:
			ctx.save()
			ctx.translate(x_offset, 0.0)
			
			# Socket number
			ctx.translate(0, height*(1 - (2*self.TEXT_ROW_HEIGHT)))
			self._draw_text( ctx
			               , self.socket_names[d]
			               , height*self.TEXT_ROW_HEIGHT
			               )
			
			# Draw the progress
			ctx.translate(0, height*self.TEXT_ROW_HEIGHT)
			self._draw_text( ctx
			               , "C%d F%d B%02d"%(c,r,s)
			               , height*self.TEXT_ROW_HEIGHT
			               )
			ctx.restore()
		
		# Draw a full-screen "paused" indicator, if paused
		if self.timing_logger is not None and self.timing_logger.paused:
			ctx.save()
			
			ctx.translate(width / 2.0, height / 2.0)
			scale = min(width, height) * 0.4
			ctx.scale(scale, scale)
			
			ctx.move_to(0, 0)
			ctx.new_sub_path()
			ctx.arc(0, 0, 1.0, 0.0, 2.0 * pi)
			ctx.set_source_rgba(0.0,0.0,0.0,0.8);
			ctx.fill_preserve()
			ctx.set_source_rgba(0.0,0.0,0.0,1.0);
			ctx.set_line_width(0.1)
			ctx.stroke()
			
			ctx.move_to(-0.25, -0.5);
			ctx.rel_line_to(0.0, 1.0);
			ctx.move_to(0.25, -0.5);
			ctx.rel_line_to(0.0, 1.0);
			ctx.set_source_rgba(0.7,0.7,0.7,1.0);
			ctx.set_line_width(0.2)
			ctx.stroke()
			
			ctx.restore()
	
	
	def _redraw(self):
		"""Redraw the GUI and display it on the screen."""
		
		# Get a new context to draw the GUI into
		height = self.tk.winfo_height()
		width = self.tk.winfo_width()
		
		surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
		ctx = cairo.Context(surface)
		
		# Render it
		self._render_gui(ctx, width, height)
		
		# Draw onto the window (note: a reference to the image is kept to keep it
		# safe from garbage collection)
		self.widget.image = ImageTk.PhotoImage(Image.frombuffer(
			"RGBA", (width, height), surface.get_data(), "raw", "BGRA", 0, 1))
		self.widget.configure(image=self.widget.image)
		self.widget.pack(expand=True, fill="both")
	
	
	def _poll_wiring_probe(self):
		"""Poll the machine's hardware to determine if the wiring is complete.
		"""
		try:
			# Check wiring conncectivity
			if self.wiring_probe is not None and self.auto_advance:
				src, dst, length = self.wires[self.cur_wire]
				
				# Check both ends of the cable
				actual_dst = self.wiring_probe.get_link_target(*src)
				actual_src = self.wiring_probe.get_link_target(*dst)
				
				advance = False
				
				if length is None:
					# We're waiting for the wire to be disconnected
					if actual_src is None and actual_dst is None:
						# Disconnected! Advance to the next wire!
						advance = True
				else:
					# We're waiting for a wire to be connected
					if actual_src == src and actual_dst == dst:
						# Connected correctly! Advance to the next wire!
						advance = True
						if self.timing_logger is not None:
							self.timing_logger.unpause()
							self.timing_logger.connection_complete()
					elif actual_dst is not None or actual_src is not None:
						# The wire was connected, but was connected incorrectly!
						if not self.connected_incorrectly:
							self._tts_speak("Wire inserted incorrectly.")
							if self.timing_logger is not None:
								self.timing_logger.unpause()
								self.timing_logger.connection_error()
						self.connected_incorrectly = True
					else:
						# No wire is connected
						self.connected_incorrectly = False
					
				# Actually advance, as required
				if advance and self.cur_wire != len(self.wires) - 1:
					self.go_to_wire(self.cur_wire + 1)
					self._redraw()
		except:
			# Fail gracefully...
			print(traceback.format_exc())
		
		# Schedule next poll
		self.tk.after(InteractiveWiringGuide.POLL_INTERVAL_MS, self._poll_wiring_probe)
	
	
	def _on_next(self, event):
		"""Advance to the next wire."""
		self.go_to_wire((self.cur_wire + 1) % len(self.wires))
		self._redraw()
	
	
	def _on_prev(self, event):
		"""Retreat to the previous wire."""
		self.go_to_wire((self.cur_wire - 1) % len(self.wires))
		self._redraw()
	
	
	def _on_first(self, event):
		"""Go back to the first wire."""
		self.go_to_wire(0)
		self._redraw()
	
	
	def _on_last(self, event):
		"""Go to the last first wire."""
		self.go_to_wire(len(self.wires)-1)
		self._redraw()
	
	
	def _on_skip_next(self, event):
		"""Advance rapidly forward through the wires."""
		self.go_to_wire((self.cur_wire + 25) % len(self.wires))
		self._redraw()
	
	
	def _on_skip_prev(self, event):
		"""Retreat rapidly backward through the wires."""
		self.go_to_wire((self.cur_wire - 25) % len(self.wires))
		self._redraw()
	
	
	def _on_tts_toggle(self, event):
		"""Toggle whether Text-to-Speech is enabled."""
		self.use_tts = not self.use_tts
		if self.use_tts:
			self._tts_speak("Text to speech enabled.")
		else:
			self._tts_speak("Text to speech disabled.")
	
	
	def _on_auto_advance_toggle(self, event):
		"""Toggle whether auto-advance is enabled."""
		if self.wiring_probe is not None:
			self.auto_advance = not self.auto_advance
			if self.auto_advance:
				self._tts_speak("Auto advance enabled.")
			else:
				self._tts_speak("Auto advance disabled.")
		else:
			self._tts_speak("Auto advance not supported.")
	
	
	def _on_pause(self, event):
		"""Toggle whether timings are being recorded."""
		if self.timing_logger is not None:
			if self.timing_logger.paused:
				self.timing_logger.unpause()
			else:
				self.timing_logger.pause()
		self._redraw()
	
	
	def _on_resize(self, event):
		"""Window has been resized, trigger a redraw."""
		new_size = (event.width, event.height)
		
		if self._old_size != new_size:
			self._old_size = new_size
			self._redraw()
	
	
	def _on_close(self, event=None):
		"""The window has been closed."""
		if self.timing_logger is not None:
			self.timing_logger.logging_stopped()
		
		# Turn off LEDs before leaving
		self.set_leds(self.cur_wire, False)
		self.tk.destroy()
	
	
	def mainloop(self):  # pragma: no cover
		"""Start the interactive wiring guide GUI. Returns when the window is
		closed."""
		return self.tk.mainloop()



#!/usr/bin/env python

"""
Coordinate systems which may be used. Coordinates can be added, subtracted, and
their magnitudes taken using .magnitude().
"""

from collections import namedtuple

################################################################################
# Base Classes (Internal Use Only)
################################################################################

class _ElementwiseCoordsMixin(object):
	"""
	Support for common operations on coordinates for which simple element-wise
	operators are adequate.
	"""
	
	def __add__(self, other):
		"""
		Performs element-wise subtraction.
		"""
		assert(len(self) == len(other))
		return type(self)(*(a+b for (a,b) in zip(self,other)))
	
	
	def __sub__(self, other):
		"""
		Performs element-wise subtraction.
		"""
		assert(len(self) == len(other))
		return type(self)(*(a-b for (a,b) in zip(self,other)))
	
	
	def __abs__(self):
		"""
		Element-wise absolute.
		"""
		return type(self)(*(abs(v) for v in self))
	
	
	def __repr__(self):
		return "%s(%s)"%(
			type(self).__name__,
			", ".join(map(repr, self))
		)



class _HexCoordsMixin(_ElementwiseCoordsMixin):
	"""
	Support for common operations on hexagonal coordinates.
	"""
	
	def __init__(self):
		_ElementwiseCoordsMixin.__init__(self)
	
	
	def magnitude(self):
		"""
		Magnitude
		"""
		# Pad to a 3-field value if 2D version.
		v = (list(self) + [0])[:3]
		from spinner import topology
		return topology.manhattan(topology.to_shortest_path(v))



class _CartesianCoordsMixin(_ElementwiseCoordsMixin):
	"""
	Support for common operations on Cartesian coordinates.
	"""
	
	def __init__(self, *args, **kwargs):
		_ElementwiseCoordsMixin.__init__(self)
	
	
	def to_positive(self):
		"""
		Return a positive-only version of the coordinate.
		"""
		return type(self)(*(abs(v) for v in self))
	
	
	def magnitude(self):
		"""
		Magnitude (Euclidean distance)
		"""
		from spinner import topology
		return topology.euclidean(self)



_HexagonalTuple = namedtuple("_HexagonalTuple", ["x","y","z"])
_Hexagonal2DTuple = namedtuple("_Hexagonal2DTuple", ["x","y"])

_Cartesian2DTuple = namedtuple("_Cartesian2DTuple", ["x","y"])
_Cartesian3DTuple = namedtuple("_Cartesian3DTuple", ["x","y","z"])

_CabinetTuple = namedtuple("_CabinetTuple", ["cabinet","frame","board"])




################################################################################
# Front-end Classes
################################################################################

"""
Hexagonal coordinate system conventionally used when working with a hexagonal
mesh such a SpiNNaker nodes or boards.

Note: The three axes are non-orthogonal and so various strange things can happen
when working with such schemes. See the topology module for various useful
functions.
"""
class Hexagonal(_HexCoordsMixin, _HexagonalTuple):
	def __init__(self, *args, **kwargs):
		_HexCoordsMixin.__init__(self)


"""
Special case of Hexagonal. Represents the Hexagonal() value with z fixed as 0.
"""
class Hexagonal2D(_HexCoordsMixin , _Hexagonal2DTuple):
	def __init__(self, *args, **kwargs):
		_HexCoordsMixin.__init__(self)


"""
Cartesian coordinates in either 2D or 3D space.
"""
class Cartesian2D(_CartesianCoordsMixin, _Cartesian2DTuple):
	def __init__(self, *args, **kwargs):
		_CartesianCoordsMixin.__init__(self)

class Cartesian3D(_CartesianCoordsMixin, _Cartesian3DTuple):
	def __init__(self, *args, **kwargs):
		_CartesianCoordsMixin.__init__(self)

"""
Logical coordinates for locations in a series of cabinets containing frames
containing boards, like so::
	
	          2             1                0
	Cabinet --+-------------+----------------+
	          |             |                |
	+-------------+  +-------------+  +-------------+    Frame
	|             |  |             |  |             |      |
	| +---------+ |  | +---------+ |  | +---------+ |      |
	| | : : : : | |  | | : : : : | |  | | : : : : |--------+ 0
	| | : : : : | |  | | : : : : | |  | | : : : : | |      |
	| +---------+ |  | +---------+ |  | +---------+ |      |
	| | : : : :#| |  | | : : : : | |  | | : : : : |--------+ 1
	| | : : : :#| |  | | : : : : | |  | | : : : : | |      |
	| +---------+ |  | +---------+ |  | +---------+ |      |
	| | : : : : | |  | | : : : : | |  | | : : : : |--------+ 2
	| | : : : : | |  | | : : : : | |  | | : : : : | |      |
	| +---------+ |  | +---------+ |  | +---------+ |      |
	| | : : : : | |  | | : : : : | |  | | : : : : |--------+ 3
	| | : : : : | |  | | : : : : | |  | | : : : : | |
	| +---------+ |  | +|-|-|-|-|+ |  | +---------+ |
	|             |  |  | | | | |  |  |             |
	+-------------+  +--|-|-|-|-|--+  +-------------+
	                    | | | | |
	         Board -----+-+-+-+-+
	                    4 3 2 1 0

In this example there are 3 cabinets each containing 4 frames which in turn
contain 5 boards.

Cabinets are numbered from 0 right-to-left. Frames are numbered from 0
top-to-bottom. Boards are numbered from 0 right-to-left. Therefore, the board
marked with "#" is at the coordinate (2,1,0).
"""
class Cabinet(_ElementwiseCoordsMixin, _CabinetTuple):
	def __init__(self, *args, **kwargs):
		_ElementwiseCoordsMixin.__init__(self)





"""Data structure which defines the physical dimensions of a set of cabinets."""

from spinner.topology import Direction

from spinner.coordinates import Cartesian3D

class Cabinet(object):
	"""Data structure which defines the physical dimensions of a set of
	cabinets.
	
	All coordinates and vectors are relative to the left-top-front corner of the
	machine.
	"""
	
	def __init__(self,
	             board_dimensions,
	             board_wire_offset_south_west, board_wire_offset_north_east,
	             board_wire_offset_east, board_wire_offset_west,
	             board_wire_offset_north, board_wire_offset_south,
	             inter_board_spacing,
	             boards_per_frame, frame_dimensions, frame_board_offset,
	             inter_frame_spacing,
	             frames_per_cabinet, cabinet_dimensions, cabinet_frame_offset,
	             inter_cabinet_spacing,
	             num_cabinets):
		"""Create a new Cabinet structure with the specified measurements (all in
		meters).
		
		Parameters
		----------
		board_dimensions : (w, h, d)
			Physical board dimensions in meters.
		board_wire_offset_south_west : (x, y, z)
			Physical offset of the South West connector from board left-top-front
			corner in meters. Others are similar.
		inter_board_spacing : float
			Physical spacing between each board in a frame in meters.
		boards_per_frame : int
			Number of boards in each frame.
		frame_dimensions : (w, h, d)
			Frame physical dimensions in meters.
		frame_board_offset : (x, y, z)
			Physical offset of the left-most board from the left-top-front corner of
			a frame in meters.
		inter_frame_spacing : float
			Physical spacing between frames in a cabinet in meters.
		frames_per_cabinet : int
			Number of frames per cabinet.
		cabinet_dimensions : (w, h, d)
			Cabinet physical dimensions in meters.
		cabinet_frame_offset : (x, y, z)
			Physical offset of the top frame from the left-top-front corner of a
			cabinet in meters.
		inter_cabinet_spacing : float
			Physical spacing between each cabinet in meters.
		num_cabinets : int
			Total number of cabinets in the system
		"""
		c = Cartesian3D
		
		self.board_dimensions = c(*board_dimensions)
		
		# Combine wire offsets into a dictionary
		self.board_wire_offset = {}
		self.board_wire_offset[Direction.south_west] = c(*board_wire_offset_south_west)
		self.board_wire_offset[Direction.north_east] = c(*board_wire_offset_north_east)
		self.board_wire_offset[Direction.east]       = c(*board_wire_offset_east)
		self.board_wire_offset[Direction.west]       = c(*board_wire_offset_west)
		self.board_wire_offset[Direction.north]      = c(*board_wire_offset_north)
		self.board_wire_offset[Direction.south]      = c(*board_wire_offset_south)
		
		self.inter_board_spacing = inter_board_spacing
		
		self.boards_per_frame    = boards_per_frame
		self.frame_dimensions    = c(*frame_dimensions)
		self.frame_board_offset  = c(*frame_board_offset)
		self.inter_frame_spacing = inter_frame_spacing
		
		self.frames_per_cabinet    = frames_per_cabinet
		self.cabinet_dimensions    = c(*cabinet_dimensions)
		self.cabinet_frame_offset  = c(*cabinet_frame_offset)
		self.inter_cabinet_spacing = inter_cabinet_spacing
		
		self.num_cabinets = num_cabinets
		
		# Check that all values are positive that are meant to be...
		for field in ["board_dimensions",
		              "inter_board_spacing",
		              "boards_per_frame",
		              "frame_dimensions",
		              "frame_board_offset",
		              "inter_frame_spacing",
		              "frames_per_cabinet",
		              "cabinet_dimensions",
		              "cabinet_frame_offset",
		              "inter_cabinet_spacing",
		              "num_cabinets"]:
			value = getattr(self, field)
			if isinstance(value, tuple):
				if any(v < 0.0 for v in value):
					raise ValueError("{} must be positive".format(field))
			elif value < 0.0:
				raise ValueError("{} must be positive".format(field))
		
		# Check all board wires are within the bounds of the board
		for direction in [Direction.south_west,
		                  Direction.north_east,
		                  Direction.east,
		                  Direction.west,
		                  Direction.north,
		                  Direction.south]:
			if any(v < 0.0 or v > m for (v, m)
			       in zip(self.board_wire_offset[direction], self.board_dimensions)):
				raise ValueError("{} wire must be within bounds of board".format(
					direction.name))
		
		# Check boards fit in the frame
		if any(v > m for (v, m)
		       in zip((self.frame_board_offset +
		               self.get_dimensions(boards=self.boards_per_frame)),
		              self.frame_dimensions)):
			raise ValueError("boards must be within bounds of a frame")
		
		# Check frames fit in the cabinet
		if any(v > m for (v, m)
		       in zip((self.cabinet_frame_offset +
		               self.get_dimensions(frames=self.frames_per_cabinet)),
		              self.cabinet_dimensions)):
			raise ValueError("frames must be within bounds of a cabinet")
	
	
	def get_position(self, cabinet, frame=None, board=None, wire=None):
		"""
		Get the physical position of a given cabinet, frame, board or wire (i.e. the
		left-top-front corner).
		"""
		
		assert 0 <= cabinet < self.num_cabinets
		
		# Note: cabinets go right-to-left(!)
		pos = Cartesian3D(0, 0, 0)
		pos += Cartesian3D(((self.cabinet_dimensions.x + self.inter_cabinet_spacing) *
		                    (self.num_cabinets - cabinet - 1)),
		                   0,
		                   0)
		
		if frame is None:
			assert board is None and wire is None
			return pos
		
		assert 0 <= frame < self.frames_per_cabinet
		
		pos += self.cabinet_frame_offset
		pos += Cartesian3D(0, (self.frame_dimensions.y + self.inter_frame_spacing) * frame, 0)
		
		if board is None:
			assert wire is None
			return pos
		
		assert 0 <= board < self.boards_per_frame
		
		# Note: boards go right-to-left(!)
		pos += self.frame_board_offset
		pos += Cartesian3D(((self.board_dimensions.x + self.inter_board_spacing) *
		                    (self.boards_per_frame - board - 1)),
		                   0,
		                   0)
		
		if wire is None:
			return pos
		
		pos += self.board_wire_offset[wire]
		
		return pos
	
	
	def get_position_opposite(self, cabinet, frame=None, board=None, wire=None):
		"""
		Get the physical position of the opposite-corner of a given cabinet, frame,
		board or wire (i.e. the right-bottom-back corner).
		"""
		
		pos = self.get_position(cabinet, frame, board, wire)
		
		if frame is None:
			assert board is None and wire is None
			return pos + self.cabinet_dimensions
		
		if board is None:
			assert wire is None
			return pos + self.frame_dimensions
		
		if wire is None:
			return pos + self.board_dimensions
		
		# Wires have zero size so no adjustment is necessary
		return pos
	
	
	def get_dimensions(self, cabinets=None, frames=None, boards=None):
		"""
		Get the dimensions of the specified number of the system (no arguments) or a
		specified number of cabinets, frames or boards.
		"""
		if cabinets is not None:
			assert frames is None and boards is None
			assert 0 < cabinets <= self.num_cabinets
			# Note: cabinets go in opposite directions
			return self.get_position_opposite(0) - self.get_position(cabinets-1)
		elif frames is not None:
			assert cabinets is None and boards is None
			assert 0 < frames <= self.frames_per_cabinet
			return self.get_position_opposite(0, frames-1) - self.get_position(0, 0)
		elif boards is not None:
			assert cabinets is None and frames is None
			assert 0 < boards <= self.boards_per_frame
			# Note: boards go in opposite directions
			return self.get_position_opposite(0, 0, 0) - self.get_position(0, 0, boards-1)
		else:
			return self.get_dimensions(self.num_cabinets)



"""A logger for cable installation times."""

from six import iteritems

import time
import datetime

class TimingLogger(object):
	"""A tool which logs cable installation events and their timings.
	
	This tool is designed to record detailed information chronicling the process
	of cable installation in large machines for research purposes. The tool
	outputs a CSV of events in a format suitable for processing using R or
	similar tools.
	
	The timing logger contains a state machine which is advanced by the reporting
	of various low-level events (e.g. "starting cable installation" and "cable
	installed correctly"). This state machine logs the incoming fine-grained
	event stream while also generating higher-level events.
	
	The CSV output contains many columns, some of which are only applicable to
	certain types of event. The first column, "type", defines the event type and
	is always populated. Other columns are populated dependent on the type of
	event.
	"""
	
	# An in-order list of all columns present in the CSV.
	_COLUMNS = [
		# The type of event
		"event_type",
		# The real time and date
		"realtime",
		# Time that the event occurred, in seconds since the start of logging and
		# excluding any time spent paused.
		"time",
		# Source and destination cabinet, frame, board and direction
		"sc", "sf", "sb", "sd",
		"dc", "df", "db", "dd",
		# Overall time to connect a cable correctly
		"duration",
		# Time since last attempt to connect the cable
		"attempt_duration",
		# Number of attempts made to install the current cable
		"num_attempts",
	]
	
	def __init__(self, file, add_header=True):
		"""Start logging into the specified file.
		
		Parameters
		----------
		file : file-like
			The file into which the CSV data will be written.
		add_header : bool
			Determines whether a header line defining all columns will be added to
			the file. When appending to an existing file, this should generally be
			False.
		"""
		self.file = file
		
		if add_header:
			self.file.write("{}\n".format(",".join(self._COLUMNS)))
		
		# Set by logging_started()
		self._start_time = None
		
		# Set when recording is paused and used to compensate timers for time spent
		# paused.
		self._pause_start_time = None
		
		# The current connection being installed (if any). If set, this is a
		# dictionary mapping from strings "sc", "sf", "sb", "sl", and "dc", "df",
		# "db", "dl", to the source and destination frame, cabinet, board and link
		# the connection is between.
		self._cur_connection = None
		
		# The time the current connection started being installed
		self._cur_connection_start_time = None
		
		# The time the last connection error for the current connection occurred
		# (or the connection start time if no errors have occurred.
		self._cur_connection_last_error_time = None
		
		# The number of errors made during installation of the current cable
		self._cur_connection_errors = None
	
	@property
	def paused(self):
		return self._pause_start_time is not None
	
	def _write_row(self, **columns):
		"""Write a row to the CSV."""
		# Sanity check no extra fields have made it in
		assert set(columns) <= set(self._COLUMNS)
		
		# Write out the line
		self.file.write("{}\n".format(",".join(str(columns.get(c, "NA"))
		                                       for c in self._COLUMNS)))
	
	def _now(self):
		"""Get the currently elapsed time in seconds."""
		return time.time() - self._start_time
	
	def _realtime(self):
		"""Return the current real time as a sensibly formatted string."""
		return datetime.datetime.now().isoformat()
	
	def logging_started(self):
		"""Event when logging has started."""
		self._start_time = time.time()
		self._write_row( event_type="logging_started"
		               , time=0.0
		               , realtime=self._realtime()
		               )
	
	def logging_stopped(self):
		"""Event when logging has started."""
		self._write_row( event_type="logging_stopped"
		               , time=self._now()
		               )
		self._start_time = None
	
	def connection_started(self, sc, sf, sb, sd, dc, df, db, dd):
		"""Event when a new cable has been presented for installation."""
		
		now = self._now()
		self._cur_connection_start_time = now
		self._cur_connection_last_error_time = now
		self._cur_connection_errors = 0
		
		self._cur_connection = {
			"sc": sc, "sf": sf, "sb": sb, "sd": sd.name,
			"dc": dc, "df": df, "db": db, "dd": dd.name,
		}
		
		self._write_row( event_type="connection_started"
		               , time=now
		               , realtime=self._realtime()
		               , **self._cur_connection
		               )
	
	def connection_error(self):
		"""Event when the current cable was installed incorrectly."""
		if self._cur_connection is None:
			return
		
		now = self._now()
		
		# The time since the last error (or start of connection)
		attempt_duration = now - self._cur_connection_last_error_time
		self._cur_connection_last_error_time = now
		
		self._write_row( event_type="connection_error"
		               , time=now
		               , realtime=self._realtime()
		               , attempt_duration=attempt_duration
		               , num_attempts=self._cur_connection_errors + 1
		               , **self._cur_connection
		               )
		
		self._cur_connection_errors += 1
	
	def connection_complete(self):
		"""Event when the current cable was installed correctly."""
		if self._cur_connection is None:
			return
		
		now = self._now()
		
		# The time since the start of the connection
		duration = now - self._cur_connection_start_time
		
		# The time since the last error (or start of connection)
		attempt_duration = now - self._cur_connection_last_error_time
		
		self._write_row( event_type="connection_complete"
		               , time=now
		               , realtime=self._realtime()
		               , duration=duration
		               , attempt_duration=attempt_duration
		               , num_attempts=self._cur_connection_errors + 1
		               , **self._cur_connection
		               )
		
		self._cur_connection = None
		self._cur_connection_start_time = None
		self._cur_connection_last_error_time = None
		self._cur_connection_errors = None
	
	def pause(self):
		"""Pause all timers."""
		self._pause_start_time = self._now()
	
	def unpause(self):
		"""Resume all timers."""
		if self._pause_start_time is None:
			return
		
		now = self._now()
		pause_duration = now - self._pause_start_time
		
		# Compensate for time spent paused
		self._start_time += pause_duration
		
		# Report it in the logs
		self._write_row( event_type="pause"
		               , time=now - pause_duration
		               , duration=pause_duration
		               )
		
		self._pause_start_time = None



#!/usr/bin/env python

"""Generate maps from topological positions to physical locations."""

import argparse

from spinner.diagrams.machine_map import draw_machine_map, \
	get_machine_map_aspect_ratio

from spinner.utils import folded_torus

from spinner import transforms

from spinner.scripts import arguments
from spinner.scripts.contexts import PDFContextManager, PNGContextManager


def main(args=None):
	parser = argparse.ArgumentParser(
		description="Generate visual maps from the SpiNNaker network topology to "
		            "board locations.")
	arguments.add_version_args(parser)
	arguments.add_image_args(parser)
	arguments.add_topology_args(parser)
	arguments.add_cabinet_args(parser)
	
	# Process command-line arguments
	args = parser.parse_args(args)
	(w, h), transformation, uncrinkle_direction, folds =\
		arguments.get_topology_from_args(parser, args)
	
	cabinet, num_frames = arguments.get_cabinets_from_args(parser, args)
	
	aspect_ratio = get_machine_map_aspect_ratio(w, h)
	
	output_filename, file_type, image_width, image_height =\
		arguments.get_image_from_args(parser, args, aspect_ratio)
	
	# Generate folded system
	hex_boards, folded_boards = folded_torus(w, h,
	                                         transformation,
	                                         uncrinkle_direction,
	                                         folds)
	
	# Divide into cabinets
	cabinetised_boards = transforms.cabinetise(folded_boards,
	                                           cabinet.num_cabinets,
	                                           num_frames,
	                                           cabinet.boards_per_frame)
	cabinetised_boards = transforms.remove_gaps(cabinetised_boards)
	
	# Render the image
	Context = {"png": PNGContextManager, "pdf": PDFContextManager}[file_type]
	with Context(output_filename, image_width, image_height) as ctx:
		draw_machine_map(ctx, image_width, image_height,
		                 w, h, hex_boards, cabinetised_boards)
	
	return 0


if __name__=="__main__":  # pragma: no cover
	import sys
	sys.exit(main())




"""
Tools for producing diagrams which show the network topology overlaid with board
boundaries and physical locations.
"""

import cairocffi as cairo

from math import pi, sin, cos, tan

from spinner.topology import to_xy, board_to_chip, hexagon_zero
from spinner.topology import Direction

# Margin (in chip widths) around the diagram (must be enough for labels)
MARGIN = 2

# Colours/thicknesses for drawing edges of chips
INTERNAL_EDGE_COLOUR = (0.82, 0.84, 0.81, 1.0)
N_S_EDGE_COLOUR      = (0.80, 0.00, 0.00, 1.0)
W_E_EDGE_COLOUR      = (0.30, 0.60, 0.02, 1.0)
NE_SW_EDGE_COLOUR    = (0.20, 0.40, 0.64, 1.0)

INTERNAL_EDGE_WIDTH = 0.03
N_S_EDGE_WIDTH      = 0.1
W_E_EDGE_WIDTH      = 0.1
NE_SW_EDGE_WIDTH    = 0.1

# Text label sizes/colours
BTM_LEFT_CHIP_LABEL_COLOUR = (0.33, 0.34, 0.32, 1.0)
BOARD_LABEL_COLOUR         = (0.33, 0.34, 0.32, 0.5)
COORD_LABEL_COLOUR         = (0.53, 0.54, 0.52, 1.0)

BTM_LEFT_CHIP_LABEL_SIZE = 0.6
BOARD_LABEL_SIZE = 1.1
COORD_LABEL_SIZE = 0.6

COORD_LABEL_OFFSET = 0.4

# Dot to place on the bottom left chip
BTM_LEFT_CHIP_DOT_COLOUR = BTM_LEFT_CHIP_LABEL_COLOUR
BTM_LEFT_CHIP_DOT_SIZE = 0.20


def D2R(d):
	"""
	Convert degrees to radians.
	"""
	return (float(d)/180.0)*pi


def _draw_edge(ctx, x,y, edge):
	"""
	Draw the specified single edge of a hexagon whose bottom-left corner is at
	(x,y) using whatever style is currently set.
	"""
	
	edge_num = {
		Direction.south:      0,
		Direction.east:       1,
		Direction.north_east: 2,
		Direction.north:      3,
		Direction.west:       4,
		Direction.south_west: 5,
	}[edge]
	
	# The transformation to use when drawing hexagons such that they sit on an x,y
	# grid.
	ctx.save()
	x_scale = 1.0/(2.0*sin(D2R(60)))
	y_scale = 1.0/(1.0+cos(D2R(60)))
	ctx.translate( (0.25*tan(D2R(60))) * x_scale
	             , 0.5*sin(D2R(30)) * y_scale
	             )
	ctx.move_to(x,y)
	ctx.transform(cairo.Matrix( 1.0,           0.0
	                          , -cos(D2R(60)), 1.0
	                          ))
	ctx.scale( x_scale
	         , y_scale
	         )
	
	# Draw the specified edge 
	ctx.rotate(D2R(30))
	for _ in range(edge_num):
		ctx.rotate(D2R(-60))
		ctx.rel_move_to(1.0,0.0)
	ctx.rotate(D2R(-60))
	ctx.rel_line_to(1.0,0.0)
	ctx.restore()
	
	ctx.stroke()


def _draw_text(ctx, text, align_point=0.0, size = 1.0, rgba = (0.0,0.0,0.0, 1.0)):
	"""
	Draw the desired text centered vertically around (0,0) horizontally
	"align_point" along the text's width.
	"""
	ctx.save()
	
	ctx.select_font_face("Sans")
	ctx.set_source_rgba(*rgba)
	ctx.set_font_size(size)
	x,y, w,h, _w,_h = ctx.text_extents(text)
	ctx.move_to(-x + ((1.0-w)*align_point), -y - size/2)
	ctx.show_text(text)
	
	ctx.restore()


def _draw_board(ctx, x,y, width_chips, height_chips, btm_left_chip, board_label):
	"""
	Draw a board at the given position onto the given context in a system of the
	given width/height in chips. If specified a label will be added to the
	bottom-left chip and the board as a whole. Edges of chips at the edge of the
	board will be drawn differently to indicate which HSS link they will use.
	"""
	# Draw the chips
	hexagon = set(hexagon_zero())
	max_x = max(x for (x,y) in hexagon)
	max_y = max(y for (x,y) in hexagon)
	for dx,dy in hexagon:
		northempty     = (dx+0, dy+1) not in hexagon
		southempty     = (dx+0, dy-1) not in hexagon
		southwestempty = (dx-1, dy-1) not in hexagon
		westempty      = (dx-1, dy+0) not in hexagon
		eastempty      = (dx+1, dy+0) not in hexagon
		northeastempty = (dx+1, dy+1) not in hexagon
		
		x_ =   (x+dx)%width_chips
		y_ = -((y+dy)%height_chips)
		
		if northempty and northeastempty:
			ctx.set_line_width(N_S_EDGE_WIDTH)
			ctx.set_source_rgba(*N_S_EDGE_COLOUR)
		elif northempty and westempty:
			ctx.set_line_width(W_E_EDGE_WIDTH)
			ctx.set_source_rgba(*W_E_EDGE_COLOUR)
		else:
			ctx.set_line_width(INTERNAL_EDGE_WIDTH)
			ctx.set_source_rgba(*INTERNAL_EDGE_COLOUR)
		_draw_edge(ctx, x_, y_, Direction.north)
		
		if northempty and northeastempty:
			ctx.set_line_width(N_S_EDGE_WIDTH)
			ctx.set_source_rgba(*N_S_EDGE_COLOUR)
		elif northeastempty and eastempty:
			ctx.set_line_width(NE_SW_EDGE_WIDTH)
			ctx.set_source_rgba(*NE_SW_EDGE_COLOUR)
		else:
			ctx.set_line_width(INTERNAL_EDGE_WIDTH)
			ctx.set_source_rgba(*INTERNAL_EDGE_COLOUR)
		_draw_edge(ctx, x_, y_, Direction.north_east)
		
		if eastempty and southempty:
			ctx.set_line_width(W_E_EDGE_WIDTH)
			ctx.set_source_rgba(*W_E_EDGE_COLOUR)
		elif northeastempty and eastempty:
			ctx.set_line_width(NE_SW_EDGE_WIDTH)
			ctx.set_source_rgba(*NE_SW_EDGE_COLOUR)
		else:
			ctx.set_line_width(INTERNAL_EDGE_WIDTH)
			ctx.set_source_rgba(*INTERNAL_EDGE_COLOUR)
		_draw_edge(ctx, x_, y_, Direction.east)
		
		if eastempty and southempty:
			ctx.set_line_width(W_E_EDGE_WIDTH)
			ctx.set_source_rgba(*W_E_EDGE_COLOUR)
		elif southempty and southwestempty:
			ctx.set_line_width(N_S_EDGE_WIDTH)
			ctx.set_source_rgba(*N_S_EDGE_COLOUR)
		else:
			ctx.set_line_width(INTERNAL_EDGE_WIDTH)
			ctx.set_source_rgba(*INTERNAL_EDGE_COLOUR)
		_draw_edge(ctx, x_, y_, Direction.south)
		
		if westempty and southwestempty:
			ctx.set_line_width(NE_SW_EDGE_WIDTH)
			ctx.set_source_rgba(*NE_SW_EDGE_COLOUR)
		elif southempty and southwestempty:
			ctx.set_line_width(N_S_EDGE_WIDTH)
			ctx.set_source_rgba(*N_S_EDGE_COLOUR)
		else:
			ctx.set_line_width(INTERNAL_EDGE_WIDTH)
			ctx.set_source_rgba(*INTERNAL_EDGE_COLOUR)
		_draw_edge(ctx, x_, y_, Direction.south_west)
		
		if westempty and southwestempty:
			ctx.set_line_width(NE_SW_EDGE_WIDTH)
			ctx.set_source_rgba(*NE_SW_EDGE_COLOUR)
		elif northempty and westempty:
			ctx.set_line_width(W_E_EDGE_WIDTH)
			ctx.set_source_rgba(*W_E_EDGE_COLOUR)
		else:
			ctx.set_line_width(INTERNAL_EDGE_WIDTH)
			ctx.set_source_rgba(*INTERNAL_EDGE_COLOUR)
		_draw_edge(ctx, x_, y_, Direction.west)
	
	
	# Draw the board's (0,0) chip label and dot
	ctx.save()
	ctx.translate(x+0.5, -y-0.5)
	
	# Dot
	ctx.set_source_rgba(*BTM_LEFT_CHIP_DOT_COLOUR)
	ctx.arc( 0.0,0.0
	       , BTM_LEFT_CHIP_DOT_SIZE/2.0
	       , 0.0, 2.0*pi
	       )
	ctx.fill()
	
	# Label
	ctx.translate(BTM_LEFT_CHIP_DOT_SIZE*1.5, 0.0)
	_draw_text( ctx
	          , btm_left_chip
	          , align_point = 0.0
	          , size = BTM_LEFT_CHIP_LABEL_SIZE
	          , rgba = BTM_LEFT_CHIP_LABEL_COLOUR
	          )
	ctx.restore()
	
	
	# Draw the board label
	positions = []
	
	# Position (or duplicate) appropriately depending on splitting of the board
	if x+max_x < width_chips and y+max_y < height_chips:
		# Not split
		positions.append((x+(0.5*max_x), -y-(0.5*max_y)))
	elif x+max_x > width_chips:
		# Split on the right edge
		positions.append((x+(0.5*(width_chips-x)), -y-(0.33*max_y)))
		# Split on the left edge
		positions.append(((0.5*(max_x-(width_chips-x))), -y-(0.66*max_y)))
		
		board_label = board_label.replace(" ","\n")
	elif y+max_y > height_chips:  # pragma: no branch
		# Split on top edge
		positions.append((x+(0.33*max_x), -y-(0.5*(height_chips-y))))
		# Split on bottom edge
		positions.append((x+(0.66*max_x), 0.5-(0.5*(max_y-(height_chips-y)))))
	
	for x_,y_ in positions:
		ctx.save()
		ctx.translate(x_,y_)
		lines = board_label.split("\n")
		ctx.translate(0.0, -(BOARD_LABEL_SIZE*(len(lines)/2.0)))
		for num, line in enumerate(lines):
			_draw_text( ctx
			          , line
			          , align_point = 0.5
			          , size = BOARD_LABEL_SIZE
			          , rgba = BOARD_LABEL_COLOUR
			          )
			ctx.translate(0.0, BOARD_LABEL_SIZE)
		ctx.restore()


def _draw_coords(ctx, width, height):
	"""
	Given a context and a width and height in chips, draw numbered ticks along the
	edges of the diagram.
	"""
	for num, rotation, offset, alignment in ( (height, 0.0,    -1 - COORD_LABEL_OFFSET,          1.0) # Left
	                                        , (height, 0.0,    width + COORD_LABEL_OFFSET,       0.0) # Right
	                                        , (width,  0.5*pi, COORD_LABEL_OFFSET,               0.0) # Bottom
	                                        , (width,  0.5*pi, -height - 1 - COORD_LABEL_OFFSET, 1.0) # Top
	                                        ):
		ctx.save()
		ctx.rotate(rotation)
		ctx.translate(offset, -0.5)
		for i in range(num):
			_draw_text( ctx
			          , str(i)
			          , align_point = alignment
			          , size = COORD_LABEL_SIZE
			          , rgba = COORD_LABEL_COLOUR
			          )
			ctx.translate(0.0, -1.0)
		ctx.restore()

def draw_machine_map(ctx, image_width, image_height,
                     width, height,
                     hex_boards, cabinetised_boards):
	"""
	Given a Cairo context (and the size of the rectangle to draw in), draws a map of
	the machine. Takes the width and height of the system in triads, hex_boards
	(mapping 3D hexagonal board coordinates to boards) and a cabinetised version
	of the same.
	"""
	ctx.save()
	
	# Number of chips in the system
	width_chips  = width  * 12
	height_chips = height * 12
	
	# Get the extent of the cabinets used
	max_cabinet = max(c for (b,(c,f,b)) in cabinetised_boards)
	max_frame   = max(f for (b,(c,f,b)) in cabinetised_boards)
	max_board   = max(b for (b,(c,f,b)) in cabinetised_boards)
	
	# Lookup from board to cabinet position
	b2c = dict(cabinetised_boards)
	
	# Rescale the drawing such that the diagram is rescaled to nicely fit in the
	# space given
	diagram_width  = width_chips  + (2.0*MARGIN)
	diagram_height = height_chips + (2.0*MARGIN)
	scale = min(image_width/diagram_width, image_height/diagram_height)
	ctx.translate( (image_width  - (diagram_width*scale)) / 2
	             , (image_height - (diagram_height*scale)) / 2
	             )
	ctx.scale(scale, scale)
	
	ctx.set_line_cap(cairo.LINE_CAP_ROUND)
	
	# Move to bottom-left chip
	ctx.translate(MARGIN, height_chips+MARGIN)
	
	# Draw coordinates around edge
	_draw_coords(ctx, width_chips, height_chips)
	
	# Draw each board
	for board, board_pos in hex_boards:
		x,y = to_xy(board_to_chip(board_pos))
		c,f,b = b2c[board]
		_draw_board( ctx
		           , x,y, width_chips, height_chips
		           , "(%d,%d)"%(x,y)
		           , ("C%d "%c if max_cabinet else "")
		             + ("F%d "%f if max_frame else "")
		             + ("B%d"%b)
		           )
	
	ctx.restore()


def get_machine_map_aspect_ratio(width, height):
	"""Given a system size (in triads), returns the aspect ratio of the diagram
	generated by draw_machine_map.
	"""
	diagram_width  = (width * 12)  + (2.0*MARGIN)
	diagram_height = (height * 12) + (2.0*MARGIN)
	
	return diagram_width / float(diagram_height)



#!/usr/bin/env python

"""Print basic statistics about a specified machine's topology."""

import argparse

from spinner.scripts import arguments

from spinner.scripts.markdown_gen import heading, table


def main(args=None):
	parser = argparse.ArgumentParser(
		description="Print basic topological statistics for a specified "
		            " configuration of boards.")
	arguments.add_version_args(parser)
	arguments.add_topology_args(parser)
	args = parser.parse_args(args)
	(w, h), transformation, uncrinkle_direction, folds =\
		arguments.get_topology_from_args(parser, args)
	
	out = ""
	out += heading("Topology Statistics", 1)
	out += "\n"
	out += table([["Measurement", "Value", "Unit"],
	              ["Network dimensions", "{}x{}".format(w*12, h*12), "chips"],
	              ["Board array", "{}x{}".format(w, h), "triad"],
	              ["Number of boards", 3 * w * h, ""],
	              ["Number of cables", 3 * w * h * 3, ""],
	              ["Number of chips", 3 * w * h * 48, ""],
	              ["Number of cores", 3 * w * h * 48 * 18, ""],
	             ])
	
	print(out)
	
	return 0


if __name__=="__main__":  # pragma: no cover
	import sys
	sys.exit(main())



#!/usr/bin/env python

"""
A module which generates diagrams (using Cairo) of cabinetised systems.
"""

from math import pi

import cairocffi as cairo

from six import itervalues, integer_types

from spinner.topology import Direction


def normalise_slice(val, max_val):
	"""
	Utility function.
	
	Take a value which may/may-not be a slice and cast into a slice with a
	start and stop value. If an int, produce a singleton slice, if None, produce a
	0-max slice.
	"""
	if val is None:
		return slice(0, max_val)
	elif isinstance(val, integer_types):
		return slice(val, val+1)
	else:
		assert val.step is None
		
		stop = val.stop
		if stop is None:
			stop = max_val
		
		start = val.start
		if start is None:
			start = 0
		
		return slice(start, stop)


class MachineDiagram(object):
	"""
	An object which can be configured to (quickly) generate wiring diagrams of
	large SpiNNaker machines using Cairo.
	"""
	
	# Colours of various components
	#                 R    G    B    A
	CABINET_COLOUR = (0.3, 0.3, 0.3, 1.0)
	FRAME_COLOUR   = (0.7, 0.7, 0.7, 1.0)
	BOARD_COLOUR   = (0.7, 1.0, 0.7, 1.0)
	SOCKET_COLOUR  = (0.5, 0.5, 0.5, 1.0)
	
	def __init__(self, cabinet):
		"""
		cabinet_system is a spinner.cabinet.Cabinet describing the physical parameters
		of the machine in question.
		"""
		self.cabinet = cabinet
		
		# Give up if any connectors are on the back
		if any(c.z != 0.0 for c in itervalues(self.cabinet.board_wire_offset)):
			raise NotImplementedError(
				"Connectors on multiple z-planes not currently supported.")
		
		# Set of wires to draw as a list [ ((c,f,b,d) , (c,f,b,d) , rgba, width),
		# ...] specifying source and destination cabinets, frame, board and
		# directions.
		self.wires = []
		
		# Set of highlight rectangles to draw into the image before wires are added.
		# A list [(x,y, w,h, rgba, width), ...].
		self.highlights = []
		
		# Set of labels to add to the diagram.
		# A list [(str, x,y, angle, size, alignment, rgba), ...]
		self.labels = []
	
	
	def add_wire( self
	            , src , dst
	            , rgba = (0.0, 0.0, 0.0, 1.0)
	            , width = 0.002
	            ):
		"""
		Cause the listed wire to be drawn from src and to dst where these arguments
		are (cabinet, frame, board, direction) tuples.
		
		rgba is a tuple (red, green, blue, alpha) specifiying how the wire will
		appear in the image.
		
		width is a tuple specifiying how thick the wire should be drawn (in meters).
		"""
		self.wires.append((src, dst, rgba, width))
	
	
	def add_highlight(self, cabinet, frame=None, board=None, direction=None,
	                  rgba=(1.0,0.0,0.0,1.0), width = 0.01):
		"""
		Highlight a specific cabinet/frame/board/socket.
		"""
		if direction is not None:
			assert frame is not None and board is not None
			x, y, _ = self.cabinet.get_position(cabinet, frame, board, direction)
			w, h = self._socket_dimensions
			x -= w/2.0
			y -= h/2.0
		elif board is not None:
			assert frame is not None
			x, y, _ = self.cabinet.get_position(cabinet, frame, board)
			w, h, _ = self.cabinet.board_dimensions
		elif frame is not None:
			x, y, _ = self.cabinet.get_position(cabinet, frame)
			w, h, _ = self.cabinet.frame_dimensions
		else:
			x, y, _ = self.cabinet.get_position(cabinet)
			w, h, _ = self.cabinet.cabinet_dimensions
	
		self.highlights.append((x,y, w,h, rgba, width))
	
	
	def add_label(self, label, cabinet, frame=None, board=None,
	              socket=None, rgba=(0.0,0.0,0.0,1.0)):
		"""
		Add a textual label to a specific cabinet/frame/board/socket.
		"""
		if socket is not None:
			assert board is not None
			assert frame is not None
			# Render text going bottom-to-top in the center of the board
			x, y, _ = self.cabinet.get_position(cabinet, frame, board, socket)
			w, h = self._socket_dimensions
			
			alignment = 0.5
			size = w * 1.0
			angle = -pi/2.0
		elif board is not None:
			assert frame is not None
			# Render text going bottom-to-top in the center of the board
			x, y, _ = self.cabinet.get_position(cabinet, frame, board)
			w, h, _ = self.cabinet.board_dimensions
			
			alignment = 0.0
			size = w * 0.8
			x += w/2.0
			y += h*0.99
			angle = -pi/2.0
		elif frame is not None:
			# Render text going left-to-right in the space at the left-hand-side of
			# the frame before the boards start.
			
			x, y, _ = self.cabinet.get_position(cabinet, frame)
			w = self.cabinet.frame_board_offset.x
			h = self.cabinet.frame_dimensions.y
			
			alignment = 0.5
			size = w * 0.8
			x += w/2.0
			y += h/2.0
			angle = 0.0
		else:
			# Render text going left-to-right in the space at the bottom of each
			# cabinet.
			x = self.cabinet.get_position(cabinet).x
			y = self.cabinet.get_position_opposite(
				cabinet, self.cabinet.frames_per_cabinet-1).y
			w = self.cabinet.cabinet_dimensions.x
			h = self.cabinet.get_position_opposite(cabinet).y - y
			
			alignment = 0.5
			size = h * 0.8
			x += w/2.0
			y += h/2.0
			angle = 0.0
	
		self.labels.append((str(label), x,y, angle, size, alignment, rgba))
	
	
	@property
	def _socket_dimensions(self):
		"""
		As a heuristic, draw sockets as rectangles half the width of the board.
		"""
		sw = self.cabinet.board_dimensions.x / 2.0
		sh = sw * 2.0
		
		return (sw, sh)
	
	
	def _draw_board(self, ctx, cabinet_num, frame_num, board_num):
		"""
		Draw a single board. Assumes the top-left corner of the frame to
		draw is at (0,0).
		"""
		# Draw the board
		x, y, _ = self.cabinet.get_position(cabinet_num, frame_num, board_num)
		w, h, _ = self.cabinet.board_dimensions
		
		ctx.save()
		ctx.rectangle(x,y,w,h)
		ctx.set_source_rgba(*self.BOARD_COLOUR)
		ctx.fill()
		ctx.restore()
		
		# Draw the sockets
		ctx.save()
		ctx.set_source_rgba(*self.SOCKET_COLOUR)
		for direction in Direction:
			# Centre position of each socket
			sx, sy, _ = self.cabinet.get_position(cabinet_num, frame_num, board_num,
			                                      direction)
			
			# As a heuristic, draw sockets as rectangles half the width of the board.
			sw, sh = self._socket_dimensions
			sx -= sw/2.0
			sy -= sh/2.0
			
			ctx.rectangle(sx, sy, sw, sh)
			ctx.fill()
		ctx.restore()
		
	
	def _draw_frame(self, ctx, cabinet_num, frame_num, boards=None):
		"""
		Draw the specified frame.
		"""
		x, y, _ = self.cabinet.get_position(cabinet_num, frame_num)
		w, h, _ = self.cabinet.frame_dimensions
		
		ctx.save()
		ctx.rectangle(x,y,w,h)
		ctx.set_source_rgba(*self.FRAME_COLOUR)
		ctx.fill()
		ctx.restore()
		
		boards = normalise_slice(boards, self.cabinet.boards_per_frame)
		for board_num in range(boards.start, boards.stop):
			self._draw_board(ctx, cabinet_num, frame_num, board_num)
	
	
	def _draw_cabinet(self, ctx, cabinet_num, frames=None, boards=None):
		"""
		Draw the specified cabinet.
		"""
		x, y, _ = self.cabinet.get_position(cabinet_num)
		w, h, _ = self.cabinet.cabinet_dimensions
		
		ctx.save()
		ctx.rectangle(x,y,w,h)
		ctx.set_source_rgba(*self.CABINET_COLOUR)
		ctx.fill()
		ctx.restore()
		
		frames = normalise_slice(frames, self.cabinet.frames_per_cabinet)
		for frame_num in range(frames.start, frames.stop):
			self._draw_frame(ctx, cabinet_num, frame_num, boards)
	
	
	def _draw_system(self, ctx, cabinets=None, frames=None, boards=None):
		"""
		Draw every cabinet in the system.
		"""
		cabinets = normalise_slice(cabinets, self.cabinet.num_cabinets)
		for cabinet_num in range(cabinets.start, cabinets.stop):
			self._draw_cabinet(ctx, cabinet_num, frames, boards)
	
	
	def _draw_labels(self, ctx):
		"""
		Draw any labels rectangles specified using the add_label method.
		"""
		for string, x,y, angle, size, alignment, rgba in self.labels:
			ctx.save()
			
			# Set font properties
			ctx.select_font_face("Sans")
			ctx.set_source_rgba(*rgba)
			ctx.set_font_size(size)
			
			# Move into position
			ctx.move_to(x, y)
			ctx.rotate(angle)
			
			# Center the text at the current position
			x,y, w,h, _w,_h = ctx.text_extents(string)
			ctx.rel_move_to(-x -w*alignment, h*0.5)
			
			# Render the string
			ctx.show_text(string)
			
			ctx.restore()
	
	
	def _draw_highlights(self, ctx):
		"""
		Draw any highlight rectangles specified using the add_highlight method.
		"""
		for x,y, w,h, rgba, width in self.highlights:
			ctx.rectangle(x,y, w,h)
			ctx.set_source_rgba(*rgba)
			ctx.set_line_width(width)
			ctx.stroke()
	
	
	def _draw_wires(self, ctx):
		"""
		Draw all wires in the system specified by add_wire.
		"""
		for src, dst, rgba, width in self.wires:
			ctx.set_line_cap(cairo.LINE_CAP_ROUND)
			ctx.move_to(*self.cabinet.get_position(*src)[:2])
			ctx.line_to(*self.cabinet.get_position(*dst)[:2])
			ctx.set_source_rgba(*rgba)
			ctx.set_line_width(width)
			ctx.stroke()
	
	
	def draw(self, ctx, max_width, max_height,
	         cabinet=None, frame=None, board=None,
	         cabinet_mask=None, frame_mask=None, board_mask=None):
		"""
		Draw the diagram onto the supplied context at the sizes defined. Optionally
		zoom the diagram to a specific cabinet, frame or board.
		
		ctx is a Cairo drawing context onto which the diagram will be drawn
		
		max_width and max_height are the maximum dimensions of the context which
		should be used.
		
		cabinet, frame and board define a desired part of the system to focus on.
		The last to be specified may be an integer or a slice.
		
		the cabinet_mask, frame_mask and board_mask specify the ranges of cabinets,
		frames and boards which should actually be drawn. (Note: wires, highlights
		and labels going to any boards outside these ranges will still be drawn.)
		"""
		ctx.save()
		
		# Produce a slice equivalent of each argument which is either the specified
		# slice or a slice which selects everything.
		cabinets = normalise_slice(cabinet, self.cabinet.num_cabinets)
		frames = normalise_slice(frame, self.cabinet.frames_per_cabinet)
		boards = normalise_slice(board, self.cabinet.boards_per_frame)
		
		# Work out the bounding box for the selection specified
		if board is not None:
			assert cabinet is not None and frame is not None
			assert isinstance(cabinet, integer_types)
			assert isinstance(frame, integer_types)
			
			focus_x, focus_y, _ =\
				self.cabinet.get_position(cabinet, frame, boards.stop - 1)
			focus_width, focus_height, _  =\
  			self.cabinet.get_dimensions(boards=boards.stop-boards.start)
		elif frame is not None:
			assert cabinet is not None
			assert isinstance(cabinet, integer_types)
			
			focus_x, focus_y, _ =\
				self.cabinet.get_position(cabinet, frames.start)
			focus_width, focus_height, _  =\
				self.cabinet.get_dimensions(frames=frames.stop-frames.start)
		elif cabinet is not None:
			focus_x, focus_y, _ =\
				self.cabinet.get_position(cabinets.stop-1)
			focus_width, focus_height, _  =\
				self.cabinet.get_dimensions(cabinets=cabinets.stop-cabinets.start)
		else:
			# If nothing specified, show everything
			focus_x, focus_y = 0, 0
			focus_width, focus_height, _ = self.cabinet.get_dimensions()
		
		# Rescale the drawing such that the system fits the space as well as
		# possible
		fit_to_width_scale  = max_width/focus_width
		fit_to_height_scale = max_height/focus_height
		scale = min(fit_to_width_scale, fit_to_height_scale)
		ctx.translate( (max_width  - (focus_width*scale)) / 2
		             , (max_height - (focus_height*scale)) / 2
		             )
		ctx.scale(scale, scale)
		ctx.translate(-focus_x, -focus_y)
		
		# Draw the specified subset of the system
		self._draw_system(ctx,
		                  normalise_slice(cabinet_mask, self.cabinet.num_cabinets),
		                  normalise_slice(frame_mask, self.cabinet.frames_per_cabinet),
		                  normalise_slice(board_mask, self.cabinet.boards_per_frame))
		
		
		# Draw extra details
		self._draw_wires(ctx)
		self._draw_labels(ctx)
		self._draw_highlights(ctx)
		
		ctx.restore()



"""Python context managers which deal with the boilerplate for working with
Cairo PNG and PDF surfaces.
"""

import cairocffi as cairo


class PDFContextManager(object):
	"""A context manager which creates a Cairo PDF context."""
	
	# Number of points in a millimetre.
	PT_PER_MM = 2.83464567
	
	def __init__(self, filename, width, height):
		"""Context manager for a PDF context with the specified filename and with
		all units given in mm.
		"""
		self.filename = filename
		self.width = width
		self.height = height
	
	def __enter__(self):
		surface = cairo.PDFSurface(self.filename,
		                           self.width * self.PT_PER_MM,
		                           self.height * self.PT_PER_MM)
		self.ctx = cairo.Context(surface)
		# Make the base unit mm.
		self.ctx.scale(self.PT_PER_MM, self.PT_PER_MM)
		return self.ctx
	
	def __exit__(self, type, value, traceback):
		self.ctx.show_page()


class PNGContextManager(object):
	"""A context manager which creates a Cairo PNG context."""
	
	def __init__(self, filename, width, height):
		"""Context manager for a PDF context with the specified filename and with
		all units given in pixels.
		"""
		self.filename = filename
		self.width = width
		self.height = height
	
	def __enter__(self):
		self.surface = cairo.ImageSurface(cairo.FORMAT_ARGB32,
		                                  self.width,
		                                  self.height)
		ctx = cairo.Context(self.surface)
		return ctx
	
	def __exit__(self, type, value, traceback):
		self.surface.write_to_png(self.filename)





#!/usr/bin/env python

"""
A tool which checks that a target SpiNNaker machine's wiring matches the
specifications generated by SpiNNer.
"""

import sys

import argparse

from spinner.utils import folded_torus

from spinner import transforms
from spinner import plan
from spinner import probe

from spinner.scripts import arguments

from rig.machine_control import BMPController


def main(args=None):
	parser = argparse.ArgumentParser(
		description="Validate the wiring of a SpiNNaker system.")
	arguments.add_version_args(parser)
	
	parser.add_argument("--verbose", "-v", action="store_true", default=False,
	                    help="list all incorrect and missing wires")
	
	arguments.add_topology_args(parser)
	arguments.add_cabinet_args(parser)
	arguments.add_bmp_args(parser)
	
	# Process command-line arguments
	args = parser.parse_args(args)
	(w, h), transformation, uncrinkle_direction, folds =\
		arguments.get_topology_from_args(parser, args)
	
	cabinet, num_frames = arguments.get_cabinets_from_args(parser, args)
	
	bmp_ips = arguments.get_bmps_from_args(parser, args,
	                                       cabinet.num_cabinets,
	                                       num_frames)
	
	if len(bmp_ips) == 0:
		parser.error("BMP host names must be provided for every frame.")
	
	# Generate folded system
	hex_boards, folded_boards = folded_torus(w, h,
	                                         transformation,
	                                         uncrinkle_direction,
	                                         folds)
	
	# Divide into cabinets
	cabinetised_boards = transforms.cabinetise(folded_boards,
	                                           cabinet.num_cabinets,
	                                           num_frames,
	                                           cabinet.boards_per_frame)
	cabinetised_boards = transforms.remove_gaps(cabinetised_boards)
	
	# Generate list of wires
	wires = plan.enumerate_wires(cabinetised_boards)
	
	# Set up the wiring probe
	bmp_controller = BMPController(bmp_ips)
	if cabinet.num_cabinets == 1 and num_frames == 1:
		num_boards = 3 * w * h
	else:
		num_boards = cabinet.boards_per_frame
	wiring_probe = probe.WiringProbe(bmp_controller,
	                                 cabinet.num_cabinets,
	                                 num_frames,
	                                 num_boards)
	
	
	# Check for the presence of every wire
	missing = []
	b2c = dict(cabinetised_boards)
	for ((src_board, src_direction), (dst_board, dst_direction)) in wires:
		src = tuple(list(b2c[src_board]) + [src_direction])
		dst = tuple(list(b2c[dst_board]) + [dst_direction])
		actual_dst = wiring_probe.get_link_target(*src)
		actual_src = wiring_probe.get_link_target(*dst)
		
		if actual_dst != dst or actual_src != src:
			missing.append((src, dst))
	
	if missing:
		sys.stderr.write("{} wires missing or erroneously connected.\n".format(
			len(missing), len(wires)))
		
		if args.verbose:
			for src, dst in missing:
				print("C:{} F:{} B:{} {} <--> C:{} F:{} B:{} {}".format(
					src[0], src[1], src[2], src[3].name.replace("_", " "),
					dst[0], dst[1], dst[2], dst[3].name.replace("_", " ")))
		else:
			print("Add --verbose for a complete list.")
		
		return -1
	else:
		sys.stderr.write("All {} wires correctly connected.\n".format(len(wires)))
		return 0


if __name__=="__main__":  # pragma: no cover
	import sys
	sys.exit(main())



#!/usr/bin/env python

"""
A proxy server which enables multiple interactive wiring sessions to interact
with the same SpiNNaker machine.
"""

import argparse

import logging

from spinner.scripts import arguments

from spinner.probe import WiringProbe

from spinner.proxy import ProxyServer, DEFAULT_PORT

from rig.machine_control import BMPController


def main(args=None):
	parser = argparse.ArgumentParser(
		description="Start a proxy server to enable multiple interactive wiring "
		            "sessions to interact with the same SpiNNaker machine.")
	arguments.add_version_args(parser)
	
	parser.add_argument("--host", "-H", type=str, default="",
	                    help="Host interface to listen on (default: any)")
	
	parser.add_argument("--port", "-p", type=int, default=DEFAULT_PORT,
	                    help="Port listen on (default: %(default)d)")
	
	parser.add_argument("--verbose", "-v", action="count", default=0,
	                    help="Increase verbosity.")
	
	arguments.add_topology_args(parser)
	arguments.add_cabinet_args(parser)
	arguments.add_bmp_args(parser)
	
	# Process command-line arguments
	args = parser.parse_args(args)
	(w, h), transformation, uncrinkle_direction, folds =\
		arguments.get_topology_from_args(parser, args)
	
	cabinet, num_frames = arguments.get_cabinets_from_args(parser, args)
	
	bmp_ips = arguments.get_bmps_from_args(parser, args,
	                                       cabinet.num_cabinets,
	                                       num_frames)
	
	if cabinet.num_cabinets == num_frames == 1:
		num_boards = 3 * w * h
	else:
		num_boards = cabinet.boards_per_frame
	
	# Set verbosity level
	if args.verbose == 1:
		logging.basicConfig(level=logging.INFO)
	elif args.verbose >= 2:
		logging.basicConfig(level=logging.DEBUG)
	
	# Create a BMP connection
	if len(bmp_ips) == 0:
		parser.error("All BMPs must be supplied using --bmp")
	bmp_controller = BMPController(bmp_ips)
	
	# Create a wiring probe
	wiring_probe = WiringProbe(bmp_controller,
	                           cabinet.num_cabinets,
	                           num_frames,
	                           num_boards)
	
	proxy_server = ProxyServer(bmp_controller, wiring_probe,
	                           args.host, args.port)
	
	print("Proxy server starting...")
	proxy_server.main()
	
	return 0


if __name__=="__main__":  # pragma: no cover
	import sys
	sys.exit(main())




#!/usr/bin/env python

"""
An interactive graphical program for guiding a user through the process of
wiring a SpiNNaker machine.
"""

import sys

import argparse

import os.path

from spinner.utils import folded_torus

from spinner import transforms

from spinner.scripts import arguments

from spinner.plan import generate_wiring_plan, flatten_wiring_plan

from spinner.diagrams.interactive_wiring_guide import InteractiveWiringGuide

from spinner.probe import WiringProbe

from spinner.proxy import ProxyClient

from spinner.timing_logger import TimingLogger

from rig.machine_control import BMPController


def main(args=None):
	parser = argparse.ArgumentParser(
		description="Interactively guide the user through the process of wiring up a "
		            "SpiNNaker machine.")
	arguments.add_version_args(parser)
	
	parser.add_argument("--no-tts", action="store_true", default=False,
	                    help="disable text-to-speech announcements of wiring "
	                         "steps")
	
	parser.add_argument("--no-auto-advance", action="store_true", default=False,
	                    help="disable auto-advancing through wiring steps")
	
	parser.add_argument("--fix", action="store_true", default=False,
	                    help="detect errors in existing wiring and just show "
	                         "corrective steps")
	
	parser.add_argument("--log", type=str, metavar="LOGFILE",
	                    help="record the times at which each cable is installed")
	
	arguments.add_topology_args(parser)
	arguments.add_cabinet_args(parser)
	arguments.add_wire_length_args(parser)
	arguments.add_bmp_args(parser)
	arguments.add_proxy_args(parser)
	arguments.add_subset_args(parser)
	
	# Process command-line arguments
	args = parser.parse_args(args)
	(w, h), transformation, uncrinkle_direction, folds =\
		arguments.get_topology_from_args(parser, args)
	
	cabinet, num_frames = arguments.get_cabinets_from_args(parser, args)
	
	wire_lengths, min_slack = arguments.get_wire_lengths_from_args(
		parser, args, mandatory=True)
	
	bmp_ips = arguments.get_bmps_from_args(parser, args,
	                                       cabinet.num_cabinets,
	                                       num_frames)
	
	proxy_host_port = arguments.get_proxy_from_args(parser, args)
	
	wire_filter = arguments.get_subset_from_args(parser, args)
	
	if cabinet.num_cabinets == num_frames == 1:
		num_boards = 3 * w * h
	else:
		num_boards = cabinet.boards_per_frame
	
	# Generate folded system
	hex_boards, folded_boards = folded_torus(w, h,
	                                         transformation,
	                                         uncrinkle_direction,
	                                         folds)
	
	# Divide into cabinets
	cabinetised_boards = transforms.cabinetise(folded_boards,
	                                           cabinet.num_cabinets,
	                                           num_frames,
	                                           cabinet.boards_per_frame)
	cabinetised_boards = transforms.remove_gaps(cabinetised_boards)
	physical_boards = transforms.cabinet_to_physical(cabinetised_boards, cabinet)
	
	# Focus on only the boards which are part of the system
	if cabinet.num_cabinets > 1:
		focus = [slice(0, cabinet.num_cabinets)]
	elif num_frames > 1:
		focus = [0, slice(0, num_frames)]
	else:
		focus = [0, 0, slice(0, w*h*3)]
	
	
	# Generate wiring plan
	wires_between_boards, wires_between_frames, wires_between_cabinets =\
		generate_wiring_plan(cabinetised_boards, physical_boards,
		                     cabinet.board_wire_offset, wire_lengths, min_slack)
	flat_wiring_plan = flatten_wiring_plan(wires_between_boards,
	                                       wires_between_frames,
	                                       wires_between_cabinets,
	                                       cabinet.board_wire_offset)
	
	# Create a BMP connection/wiring probe or connect to a proxy
	if proxy_host_port is None:
		if len(bmp_ips) == 0:
			if args.fix:
				parser.error("--fix requires that all BMPs be listed with --bmp")
			bmp_controller = None
			wiring_probe = None
		else:
			bmp_controller = BMPController(bmp_ips)
		
		# Create a wiring probe
		if bmp_controller is not None and (not args.no_auto_advance or args.fix):
			wiring_probe = WiringProbe(bmp_controller,
			                           cabinet.num_cabinets,
			                           num_frames,
			                           num_boards)
	else:
		# Fix is not supported since the proxy client does not recreate the
		# discover_wires method of WiringProbe.
		if args.fix:
			parser.error("--fix cannot be used with --proxy")
		
		# The proxy object provides a get_link_target and set_led method compatible
		# with those provided by bmp_controller and wiring_probe. Since these are
		# the only methods used, we use the proxy client object in place of
		# bmp_controller and wiring_probe.
		bmp_controller = wiring_probe = ProxyClient(*proxy_host_port)
	
	# Create a TimingLogger if required
	if args.log:
		if os.path.isfile(args.log):
			logfile = open(args.log, "a")
			add_header = False
		else:
			logfile = open(args.log, "w")
			add_header = True
		timing_logger = TimingLogger(logfile, add_header)
	else:
		logfile = None
		timing_logger = None
	
	# Convert wiring plan into cabinet coordinates
	b2c = dict(cabinetised_boards)
	wires = []
	for ((src_board, src_direction), (dst_board, dst_direction), wire_length) \
	    in flat_wiring_plan:
		
		sc, sf, sb = b2c[src_board]
		dc, df, db = b2c[dst_board]
		wires.append(((sc, sf, sb, src_direction),
		              (dc, df, db, dst_direction),
		              wire_length))
	
	# Filter wires according to user-specified rules
	wires = list(filter(wire_filter, wires))
	if len(wires) == 0:
		parser.error("--subset selects no wires")
	
	if not args.fix:
		# If running normally, just run through the full set of wires
		wiring_plan = wires
	else:
		# If running in fix mode, generate a list of fixes to make
		correct_wires = set((src, dst) for src, dst, length in wires)
		actual_wires = set(wiring_probe.discover_wires())
		
		to_remove = actual_wires - correct_wires
		to_add = correct_wires - actual_wires
		
		# Remove all bad wires first, then re-add good ones (note ordering now is
		# just reset to cabinets right-to-left, frames top-to-bottom and boards
		# left-to-right).
		wiring_plan = [(src, dst, None) for src, dst in sorted(to_remove)]
		for src, dst, length in wires:
			if (src, dst) in to_add:
				wiring_plan.append((src, dst, length))
		
		if len(wiring_plan) == 0:
			print("No corrections required.")
			return 0

	
	# Intialise the GUI and launch the mainloop
	ui = InteractiveWiringGuide(cabinet=cabinet,
	                            wire_lengths=wire_lengths,
	                            wires=wiring_plan,
	                            bmp_controller=bmp_controller,
	                            use_tts=not args.no_tts,
	                            focus=focus,
	                            wiring_probe=wiring_probe,
	                            auto_advance=not args.no_auto_advance,
	                            timing_logger=timing_logger)
	ui.mainloop()
	
	if logfile is not None:
		logfile.close()
	
	return 0


if __name__=="__main__":  # pragma: no cover
	import sys
	sys.exit(main())




#!/usr/bin/env python

"""
A library of functions which calculate various metrics about systems of boards
and their wiring.
"""

import math

from six import integer_types

from spinner import coordinates


def wire_lengths(boards, direction, board_wire_offset=None):
	"""
	Generate a list of as-the-crow-files wire lengths for the supplied system.
	
	board_wire_offset is an (optional) dict {direction:offset,...} where the
	offset of each wire from the left-top-front corner of a board is supplied.
	This structure can be attained from a spinner.cabinet.Cabinet object.
	"""
	b2c = dict(boards)
	
	for board, coord in boards:
		source = b2c[board]
		target = b2c[board.connection[direction]]
		
		# Up-cast the coordinates to 3D
		if len(source) == 2:
			source = coordinates.Cartesian3D(source[0], source[1], 0)
		if len(target) == 2:
			target = coordinates.Cartesian3D(target[0], target[1], 0)
		
		if board_wire_offset is not None:
			source += board_wire_offset[direction]
			target += board_wire_offset[direction.opposite]
		
		yield (source - target).magnitude()


def physical_wire_length(distance, available_wire_lengths, min_slack):
	"""
	Selects a wire length for a wire required to span a given distance with a
	certain amount of slack. Returns a tuple (selected_wire_length, slack).
	
	distance is a crow-flies distance between the sockets to be connected.
	
	available_wire_lengths is a list of available wire lengths
	
	min_slack is the minimum slack to allow.
	"""
	try:
		wire_length = min(l for l in available_wire_lengths
		                  if l >= (distance + min_slack))
		slack = wire_length - distance
		return (wire_length, slack)
	except ValueError:
		raise ValueError(
			"No wire is long enough to span a %0.3f m gap with %0.3f m of slack."%(
				distance, min_slack)) 


def wire_length_histogram(wire_lengths, min_slack, bins=10):
	"""
	Generate a histogram of physical wire lengths.
	
	wire_lengths is an iterable of crow-flies wire length values (e.g. from
	wire_lengths). Wires will be chosen for each length based on the
	physical_wire_length function.
	
	min_slack is the minimum amount of slack to provide in a wire.
	
	bins is either an int (giving the total number of bins) or a list of wire
	lengths. If the last bin supplied is too short to accomodate the largest
	input wire with sufficient slack, a ValueError will be raised.
	
	Returns a (bin_counts, bin_min_slack, bin_max_slack). bin_counts is a dictionary
	{bin: count, ...} with one entry per histogram bin. Each bin ranges from
	zero or the next-lowest bin to the specified bin. bin_min_slack and
	bin_max_slack record the range of slack for wires in each bin.
	"""
	wire_lengths = sorted(wire_lengths)
	
	# Auto-generate bins if required
	if isinstance(bins, integer_types):
		max_physical_wire_length = max(wire_lengths) + min_slack
		bins = [((n+1)/float(bins)) * float(max_physical_wire_length)
		        for n in range(bins)]
	else:
		bins = sorted(map(float, bins))
	
	# Construct the histogram
	bin_counts = {bin: 0 for bin in bins}
	bin_min_slack = {bin: bin for bin in bins}
	bin_max_slack = {bin: 0.0 for bin in bins}
	for distance in wire_lengths:
		bin, slack = physical_wire_length(distance, bins, min_slack)
		bin_counts[bin] += 1
		bin_min_slack[bin] = min(bin_min_slack[bin], slack)
		bin_max_slack[bin] = max(bin_max_slack[bin], slack)
	
	return bin_counts, bin_min_slack, bin_max_slack

def dimensions(boards):
	"""
	Return the width and height of the space occupied by the supplied set of
	boards.
	"""
	if len(boards) == 0:
		raise ValueError("Expected at least one board")
	
	return type(boards[0][1])(*(max(c[i] for b, c in boards) + 1
	                            for i in range(len(boards[0][1]))))


def count_wires(boards, direction):
	"""
	Given a cabinetised system, count the number of wires connecting between
	cabinets, between frames or within frames.
	
	direction if given restricts the count to just those directions
	
	Returns a tuple (between_cabinets, between_frames, between_boards) giving the
	counts of wires in each of the categories described above respectively.
	"""
	b2c = dict(boards)
	
	between_cabinets = 0
	between_frames = 0
	between_boards = 0
	
	for board, coord in boards:
		other = board.follow_wire(direction)
		other_coord = b2c[other]
		
		if coord.cabinet != other_coord.cabinet:
			between_cabinets += 1
		elif coord.frame != other_coord.frame:
			between_frames += 1
		else:  # if coord.board != other_coord.board:
			between_boards += 1
	
	return (between_cabinets, between_frames, between_boards)




#!/usr/bin/env python

"""Print basic statistics about a specified machine's wiring."""

import argparse

from six import iteritems, itervalues

from spinner.topology import Direction

from spinner.utils import folded_torus

from spinner import metrics
from spinner import transforms

from spinner.scripts import arguments

from spinner.scripts.markdown_gen import heading, table


def avg_wire_length_table(boards, units, wire_offsets=None):
	"""Render a table of average wire lengths for the given system."""
	wire_lengths = {d: list(metrics.wire_lengths(boards, d, wire_offsets))
	                for d in [Direction.north_east,
	                          Direction.north,
	                          Direction.west]}
	mean_wire_lengths = {d: sum(wl)/len(wl) for d, wl in iteritems(wire_lengths)}
	mean_wire_length = sum(itervalues(mean_wire_lengths))/len(mean_wire_lengths)
	max_wire_lengths = {d: max(wl) for d, wl in iteritems(wire_lengths)}
	max_wire_length = max(itervalues(max_wire_lengths))
	
	return table([["Parameter", "Value", "Unit"],
	              ["System dimensions",
	               " x ".join("%0.2f"%v for v in metrics.dimensions(boards)),
	               units],
	              ["Mean wire length", "%0.2f"%mean_wire_length, units],
	              ["  NE/SW", "  %0.2f"%mean_wire_lengths[Direction.north_east], units],
	              ["  N/S", "  %0.2f"%mean_wire_lengths[Direction.north], units],
	              ["  W/E", "  %0.2f"%mean_wire_lengths[Direction.west], units],
	              ["Maximum wire length", "%0.2f"%max_wire_length, units],
	              ["  NE/SW", "  %0.2f"%max_wire_lengths[Direction.north_east], units],
	              ["  N/S", "  %0.2f"%max_wire_lengths[Direction.north], units],
	              ["  W/E", "  %0.2f"%max_wire_lengths[Direction.west], units],
	             ])


def wire_counts_table(boards):
	"""Render a table which displays counts of wires."""
	table_data = [["Axis", "Total", "Between Cabinets", "Between Frames", "Within Frames"]]
	
	totals = [0, 0, 0, 0]
	
	for name, axis in [("NE/SW", Direction.north_east),
	                   ("N/S", Direction.north),
	                   ("W/E", Direction.west)]:
		counts = metrics.count_wires(boards, axis)
		table_data.append([name, sum(counts)] + list(counts))
		
		for col_num in range(4):
			totals[col_num] += table_data[-1][col_num + 1]
	
	table_data.append(["All"] + totals)
	
	return table(table_data)


def wire_length_table(boards, bins, min_slack,
                      wire_offsets=None, bar_length=15):
	"""Render a histogram of wire lengths in the system."""
	wire_lengths = sum((list(metrics.wire_lengths(boards, d, wire_offsets))
	                    for d in [Direction.north_east,
	                              Direction.north,
	                              Direction.west]),
	                   [])
	
	bin_counts, bin_min_slack, bin_max_slack = metrics.wire_length_histogram(
		wire_lengths, min_slack, bins)
	
	data = [("Range (meters)", "Count", "Histogram", "Min slack (meters)", "Max slack (meters)")]
	
	max_count = max(itervalues(bin_counts))
	last_bin = 0.0
	for bin in sorted(bin_counts):
		count = bin_counts[bin]
		min_slack = bin_min_slack[bin]
		max_slack = bin_max_slack[bin]
		data.append(("%0.2f < x <= %0.2f"%(last_bin, bin),
		             count,
		             "#"*((count * bar_length + max_count - 1) // max_count),
		             "%0.2f"%min_slack,
		             "%0.2f"%max_slack))
		last_bin = bin
	
	return table(data)


def main(args=None):
	parser = argparse.ArgumentParser(
		description="Print basic wiring statistics for a specified "
		            " configuration of boards.")
	arguments.add_version_args(parser)
	arguments.add_topology_args(parser)
	arguments.add_histogram_args(parser)
	arguments.add_wire_length_args(parser)
	arguments.add_cabinet_args(parser)
	
	# Process and display command-line arguments
	args = parser.parse_args(args)
	(w, h), transformation, uncrinkle_direction, folds =\
		arguments.get_topology_from_args(parser, args)
	histogram_bins = arguments.get_histogram_from_args(parser, args)
	wire_lengths, min_slack =\
		arguments.get_wire_lengths_from_args(parser, args)
	cabinet, num_frames = arguments.get_cabinets_from_args(parser, args)
	
	print(heading("Wiring Statistics", 1))
	print(heading("Folding Parameters", 2))
	print(table([["Parameter", "Value", "Unit"],
	             ["Number of boards", 3 * w * h, ""],
	             ["System dimensions", "{}x{}".format(w, h), "triads"],
	             ["Transformation", transformation, ""],
	             ["Uncrinkle Direction", uncrinkle_direction, ""],
	             ["Folds", "{}x{}".format(*folds), "pieces"],
	             ["Number of cabinets", cabinet.num_cabinets, ""],
	             ["Number of frames-per-cabinet", num_frames, ""],
	             ["Number of boards-per-frame", cabinet.boards_per_frame, ""],
	            ]))
	
	# Generate folded system and report wire-lengths after folding
	hex_boards, folded_boards = folded_torus(w, h,
	                                         transformation,
	                                         uncrinkle_direction,
	                                         folds)
	print(heading("Non-cabinetised measurements", 2))
	print(avg_wire_length_table(folded_boards, "boards"))
	
	# Divide into cabinets and report crossings
	cabinetised_boards = transforms.cabinetise(folded_boards,
	                                           cabinet.num_cabinets,
	                                           num_frames,
	                                           cabinet.boards_per_frame)
	cabinetised_boards = transforms.remove_gaps(cabinetised_boards)
	
	print(heading("Inter-cabinet/Inter-frame wiring", 2))
	print(wire_counts_table(cabinetised_boards))
	
	# Map to real, physical cabinets and measure wire lengths
	physical_boards = transforms.cabinet_to_physical(cabinetised_boards, cabinet)
	
	print(heading("Cabinetised measurements", 2))
	print("All wire lengths described in this section do not include any slack.\n")
	print(avg_wire_length_table(physical_boards, "meters",
	                            cabinet.board_wire_offset))
	
	# Generate a histogram of wire lengths
	print(heading("Wire length histogram", 2))
	print("Wire lengths are selected which include at least {} meters "
	      "of slack.\n".format(min_slack))
	print(wire_length_table(physical_boards,
	                        wire_lengths if wire_lengths else histogram_bins,
	                        min_slack,
	                        cabinet.board_wire_offset))
	
	return 0


if __name__=="__main__":  # pragma: no cover
	import sys
	sys.exit(main())




"""Functions for producing practical wiring plans."""

from six import iteritems

from collections import defaultdict

from spinner.topology import Direction

from spinner import topology
from spinner import board
from spinner import metrics


def enumerate_wires(boards):
	"""
	Takes a set of boards and enumerates the wires in the network. Returns a
	list [((src_board, src_direction), (dst_board, dst_direction)),...] in no
	particular order. Directions north, east and south_west are always "sources"
	and south, west, and north_east are always "destinations".
	"""
	
	wires = []
	for src_board, src_pos in boards:
		for src_direction in [Direction.north, Direction.east, Direction.south_west]:
			dst_board = src_board.follow_wire(src_direction)
			dst_direction = src_direction.opposite
			
			wires.append(((src_board, src_direction), (dst_board, dst_direction)))
	
	return wires


def partition_wires(wires, cabinetised_boards):
	"""
	Partition a list of wires up by whether they stay in the same frame, cabinet or
	not. Returns a tuple of two dictionaries and a list::
		
		( {(cabinet,frame): wires_between_boards,...}
		, {cabinet: wires_between_frames,...}
		, wires_between_cabinets
		)
	"""
	# Get the mapping from boards to their cabinet positions
	b2c = dict(cabinetised_boards)
	
	# {(cabinet,frame): [wire,...]}
	wires_between_boards = defaultdict(list)
	# {(cabinet): [wire,...]}
	wires_between_frames = defaultdict(list)
	# [wire,...]
	wires_between_cabinets = []
	
	for wire in wires:
		src_pos = b2c[wire[0][0]]
		dst_pos = b2c[wire[1][0]]
		
		if src_pos[0:2] == dst_pos[0:2]:
			# Same cabinet and frame
			wires_between_boards[(src_pos[0:2])].append(wire)
		elif src_pos[0] == dst_pos[0]:
			# Same cabinet
			wires_between_frames[src_pos[0]].append(wire)
		else:
			# Different cabinet
			wires_between_cabinets.append(wire)
	
	return (wires_between_boards, wires_between_frames, wires_between_cabinets)


def assign_wires(wires, physical_boards, board_wire_offset,
                 available_wire_lengths, min_slack):
	"""
	Given a list `[((src_board,src_direction),(dst_board,dst_direction)),...]`,
	sort into an order where the tightest wires are connected first. Returns::
		
		[((src_board,src_direction),(dst_board,dst_direction), wire_length),...]
	
	Where wire_length is the length of the wire assigned to that connection taken
	from the list available_wire_lengths.
	
	min_slack is the minimum amount of slack required of a cable.
	"""
	b2p = dict(physical_boards)
	d2o = board_wire_offset
	
	def assign_wire(distance):
		"""Return a tuple (wire_length, slack) shortest possible wire which coveres
		the distance while allowing the required amount of slack."""
		return metrics.physical_wire_length(distance, available_wire_lengths,
		                                    min_slack)
	
	# Augment each wire with a wire length and amount of slack
	wires = [(src, dst, assign_wire(((b2p[src[0]] + d2o[src[1]]) -
	                                 (b2p[dst[0]] + d2o[dst[1]])).magnitude()))
	          for (src, dst) in wires]
	
	# Sort the wires such that the shortest, furthest-stretched wires are
	# connected first but beyond that we move left-to-right.
	wires.sort(key=(lambda wire: (wire[2][1],  # Least-slack first first
	                              b2p[wire[0][0]].x,  # Left-most next
	                              b2p[wire[0][0]].y)))  # Top-most next
	
	# Strip out the slack and return
	return [(src, dst, length) for src, dst, (length, slack) in wires]


def generate_wiring_plan(cabinetised_boards, physical_boards,
                         board_wire_offset, available_wire_lengths,
                         min_slack):
	"""
	Get a wiring plan broken down into various stages.
	
	Takes a cabinetised torus (cabinetised_boards), the physical positions of the
	boards in space (physical_boards) and a list of available wire lengths
	(available_wire_lengths).
	
	min_slack is the minimum slack in any wire to allow.
	
	Produces a tuple of three dictionaries::
		
		(
			# Connections between boards
			{ (cabinet, frame, direction)
			  : [((src_board,src_direction),(dst_board,dst_direction), wire_length),...]
			},
			
			# Connections between frames
			{ (cabinet, direction)
			  : [((src_board,src_direction),(dst_board,dst_direction), wire_length),...]
			},
			
			# Connections between cabinets
			{ direction
			  : [((src_board,src_direction),(dst_board,dst_direction), wire_length),...]
			}
		)
	
	Which gives an ordered list of wiring connections to make for each cabinet,
	frame and axis of the system.
	"""
	b2c = dict(cabinetised_boards)
	b2p = dict(physical_boards)
	
	plan_between_boards   = {}
	plan_between_frames   = {}
	plan_between_cabinets = {}
	
	# List all the wires which exist
	wires = enumerate_wires(cabinetised_boards)
	
	# Split up wires by directon
	for direction in [Direction.north, Direction.east, Direction.south_west]:
		direction_wires = [ wire for wire in wires if wire[0][1] == direction or wire[1][1] == direction ]
		
		# Split wires up depending on whether they're within a single frame/cabinet
		# or not.
		wires_between_boards, wires_between_frames, wires_between_cabinets = \
			partition_wires(direction_wires, cabinetised_boards)
		
		for (cabinet, frame), w in iteritems(wires_between_boards):
			plan_between_boards[(cabinet, frame, direction)] = \
				assign_wires( w
				            , physical_boards
				            , board_wire_offset
				            , available_wire_lengths
				            , min_slack
				            )
		
		for cabinet, w in iteritems(wires_between_frames):
			plan_between_frames[(cabinet, direction)] = \
				assign_wires( w
				            , physical_boards
				            , board_wire_offset
				            , available_wire_lengths
				            , min_slack
				            )
		
		plan_between_cabinets[direction] = \
			assign_wires( wires_between_cabinets
			            , physical_boards
			            , board_wire_offset
			            , available_wire_lengths
			            , min_slack
			            )
	
	return (plan_between_boards, plan_between_frames, plan_between_cabinets)


def flatten_wiring_plan( wires_between_boards
                       , wires_between_frames
                       , wires_between_cabinets
                       , board_wire_offset
                       ):
	"""
	Takes the output of generate_wiring_plan and produces a flat list
	[((src_board, src_direction), (dst_board, dst_direction), wire_length), ...]
	in a sensible order.
	"""
	out = []
	
	# Wires between boards in the same frame
	cabinets = set(c for (c,f,d) in wires_between_boards)
	for cabinet in sorted(cabinets):
		frames = set(f for (c,f,d) in wires_between_boards
		               if c == cabinet
		            )
		for frame in sorted(frames):
			directions = set(d for (c,f,d) in wires_between_boards
			                   if c == cabinet and f == frame
			                )
			for direction in sorted(directions, key=(lambda d: board_wire_offset[d].y)):
				out += wires_between_boards[(cabinet,frame,direction)]
	
	# Wires between frames in the same cabinet
	cabinets = set(c for (c,d) in wires_between_frames)
	for cabinet in sorted(cabinets):
		directions = set(d for (c,d) in wires_between_frames
		                   if c == cabinet
		                )
		for direction in sorted(directions, key=(lambda d: board_wire_offset[d].y)):
			out += wires_between_frames[(cabinet,direction)]
	
	# Wires between cabinets
	directions = set(wires_between_cabinets)
	for direction in sorted(directions, key=(lambda d: board_wire_offset[d].y)):
		out += wires_between_cabinets[direction]
	
	return out



"""Standard argument parsing routines for SpiNNer scripts."""

import argparse

import re

from six import itervalues

import spinner

from spinner.topology import Direction

from spinner.utils import ideal_system_size, folded_torus, min_num_cabinets

from spinner.cabinet import Cabinet

from spinner.proxy import DEFAULT_PORT


def CabinetAction(num_levels=4, append=False):
	""""An argparse Action which accepts cabinet/frame/board/link references."""
	assert 1 <= num_levels <= 4
	
	class _CabinetAction(argparse.Action):
		
		"""Names of directions for command line arguments."""
		DIRECTION_NAMES = {d.name.replace("_", "-"): d for d in Direction}
		
		def __init__(self, *args, **kwargs):
			kwargs.setdefault("type", str)
			kwargs.setdefault("nargs", "+")
			
			metavar = ""
			if num_levels >= 4:
				metavar = " [{{{}}}{}]".format(",".join(self.DIRECTION_NAMES), metavar)
			if num_levels >= 3:
				metavar = " [BOARD{}]".format(metavar)
			if num_levels >= 2:
				metavar = " [FRAME{}]".format(metavar)
			metavar = "CABINET{}".format(metavar)
			kwargs.setdefault("metavar", (metavar, ""))
			
			argparse.Action.__init__(self, *args, **kwargs)
		
		def __call__(self, parser, namespace, values, option_string=None):
			# Fail with too many/few arguments
			if not 1 <= len(values) <= num_levels:
				parser.error("{} expects between 1 and {} values".format(
					option_string, num_levels))
			
			# Check cabinet/frame/board are integer types (and cast to int as
			# appropriate)
			for int_val, name in enumerate("CABINET FRAME BOARD".split()):
				if len(values) > int_val:
					if not values[int_val].isdigit():
						parser.error("{} value for {} must be a non-negative integer, not '{}'".format(
							option_string, name, values[int_val]))
					else:
						values[int_val] = int(values[int_val])
			
			# Convert direction into a Direction
			if len(values) >= 4:
				# Typecheck
				if not values[3] in self.DIRECTION_NAMES:
					parser.error("{} value for link must be one of {{{}}}, not {}".format(
						option_string, ",".join(self.DIRECTION_NAMES), values[3]))
				
				values[3] = self.DIRECTION_NAMES[values[3]]
			
			values = tuple(values)
			
			if append:
				if getattr(namespace, self.dest) is None:
					setattr(namespace, self.dest, [])
				getattr(namespace, self.dest).append(values)
			else:
				setattr(namespace, self.dest, values)
	
	return _CabinetAction


def add_version_args(parser):
	"""Adds a standard --version/-V incantation which prints the version number
	from spinner.__version__."""
	parser.add_argument("--version", "-V", action="version",
	                    version="%(prog)s {}".format(spinner.__version__))


def add_topology_args(parser):
	"""Add arguments for specifying SpiNNaker machine topologies and their
	folding."""
	# Require the specification of the size of the system
	topology_group = parser.add_argument_group("machine topology dimensions")
	system_size_group = topology_group.add_mutually_exclusive_group(required=True)
	system_size_group.add_argument("--num-boards", "-n", type=int, metavar="N",
	                               help="build the 'squarest' system with this "
	                                    "many boards")
	system_size_group.add_argument("--triads", "-t", type=int, nargs=2,
	                               metavar=("W", "H"),
	                               help="build a system with the specified "
	                                    "number of triads of boards in each "
	                                    "dimension (yielding 3*W*H boards)")
	
	# Arguments for specification of how the system should be folded. If left out,
	# this is done automatically
	folding_group = parser.add_argument_group("topology folding options")
	folding_group.add_argument("--transformation", "-T", choices=["shear","slice"],
	                           help="the transformation function to use from "
	                                "hexagonal torus to rectangular Cartesian "
	                                "grid (selected automatically if omitted)")
	folding_group.add_argument("--uncrinkle-direction", choices=["columns", "rows"],
	                           help="direction in which to uncrinkle the hexagonal "
	                                "mesh to form a regular grid (default: rows)")
	folding_group.add_argument("--folds", "-F", type=int, nargs=2,
	                           default=None, metavar=("X", "Y"),
	                           help="the number of pieces to fold into in each "
	                                "dimension (default: (2, 2)) ignored if "
	                                "--transformation is not given")


def get_topology_from_args(parser, args):
	"""To be used with add_topology_args.
	
	Check that the supplied arguments are valid and return corresponding
	parameters for spinner.utils.folded_torus.
	
	Returns
	-------
	((w, h), transformation, uncrinkle_direction, (x, y))
		(w, h) is the dimensions of the system in triads
		
		transformation is "slice" or "shear"
		
		uncrinkle_direction is "rows" or "columns"
		
		(x, y) gives the number of pieces to fold each dimension into
	"""
	# Extract the system dimensions
	if args.num_boards is not None:
		try:
			w, h = ideal_system_size(args.num_boards)
		except TypeError:
			parser.error("number of boards must be a multiple of three")
	else:
		w, h = args.triads
		if w <= 0 or h <= 0:
			parser.error("system dimensions must be positive and non-zero")
	
	# Fold accordingly
	transformation = "shear"
	uncrinkle_direction = "rows"
	folds = (2, 2)
	
	if args.transformation is None:
		# Work out the folding process to use by following the guidelines set out in
		# "Bringing the Hexagonal Torus Topology into the Real-World" by Heathcote
		# et. al. (unpublished at the time of writing...).
		if h == 2 * w:
			transformation = "slice"
		else:
			transformation = "shear"
		uncrinkle_direction = "rows"
		folds = (2, 2)
	else:
		transformation = args.transformation
	
	if args.folds is not None:
		folds = tuple(args.folds)
	
	if folds[0] <= 0 or folds[1] <= 0:
		parser.error("number of pieces to fold into must be at least 1")
	
	if args.uncrinkle_direction is not None:
		uncrinkle_direction = args.uncrinkle_direction
	
	return ((w, h), transformation, uncrinkle_direction, folds)


def add_cabinet_args(parser):
	"""Add arguments for specifying SpiNNaker cabinet sizes."""
	# Physical dimensions (defaults to standard rack sizes)
	board_group = parser.add_argument_group("board physical dimensions")
	board_group.add_argument("--board-dimensions", type=float, nargs=3,
	                         metavar=("W", "H", "D"),
	                         default=(0.014, 0.233, 0.240),
	                         help="physical board dimensions in meters (default: "
	                              "%(default)s)")
	for direction, default in [("south-west", (0.008, 0.013, 0.0)),
	                           ("north-east", (0.008, 0.031, 0.0)),
	                           ("east",       (0.008, 0.049, 0.0)),
	                           ("west",       (0.008, 0.067, 0.0)),
	                           ("north",      (0.008, 0.085, 0.0)),
	                           ("south",      (0.008, 0.103, 0.0))]:
		board_group.add_argument("--board-wire-offset-{}".format(direction),
		                         type=float, nargs=3, default=default,
		                         metavar=("X", "Y", "Z"),
		                         help="physical offset of the {} connector from "
		                              "board left-top-front corner in meters "
		                              "(default: %(default)s)".format(direction))
	board_group.add_argument("--inter-board-spacing", type=float,
	                         default=0.00124, metavar=("S"),
	                         help="physical spacing between each board in a "
	                              "frame in meters (default: %(default)s)")
	
	frame_group = parser.add_argument_group("frame physical dimensions")
	frame_group.add_argument("--boards-per-frame", type=int, default=24,
	                         help="number of boards per frame (default: "
	                              "%(default)s)")
	frame_group.add_argument("--frame-dimensions", type=float, nargs=3,
	                         metavar=("W", "H", "D"),
	                         default=(0.430, 0.266, 0.250),
	                         help="frame physical dimensions in meters (default: "
	                              "%(default)s)")
	frame_group.add_argument("--frame-board-offset", type=float, nargs=3,
	                         default=(0.06, 0.017, 0.0),
	                         metavar=("X", "Y", "Z"),
	                         help="physical offset of the left-top-front corner of "
	                              "the left-most board from the left-top-front "
	                              "corner of a frame in meters (default: "
	                              "%(default)s)")
	frame_group.add_argument("--inter-frame-spacing", type=float,
	                         default=0.133, metavar=("S"),
	                         help="physical spacing between frames in a "
	                              "cabinet in meters (default: %(default)s)")
	
	cabinet_group = parser.add_argument_group("cabinet physical dimensions")
	cabinet_group.add_argument("--frames-per-cabinet", type=int, default=5,
	                           help="number of frames per cabinet (default: "
	                                "%(default)s)")
	cabinet_group.add_argument("--cabinet-dimensions", type=float, nargs=3,
	                           metavar=("W", "H", "D"),
	                           default=(0.600, 2.000, 0.250),
	                           help="cabinet physical dimensions in meters (default: "
	                                "%(default)s)")
	cabinet_group.add_argument("--cabinet-frame-offset", type=float, nargs=3,
	                           default=(0.085, 0.047, 0.0),
	                           metavar=("X", "Y", "Z"),
	                           help="physical offset of the left-top-front corner "
	                                "of the top frame from the left-top-front "
	                                "corner of a cabinet in meters (default: "
	                                "%(default)s)")
	cabinet_group.add_argument("--inter-cabinet-spacing", type=float,
	                           default=0.0, metavar=("S"),
	                           help="physical spacing between each cabinet in "
	                                "meters (default: %(default)s)")
	cabinet_group.add_argument("--num-cabinets", "-c", type=int, metavar="N",
	                           help="specify how many cabinets to spread the "
	                                "system over (default: the minimum possible)")
	cabinet_group.add_argument("--num-frames", "-f", type=int, metavar="N",
                             help="when only one cabinet is required, "
                                  "specifies how many frames within that "
                                  "cabinet the system should be spread "
                                  "across (default: the minimum possible)")


def get_cabinets_from_args(parser, args):
	"""For use with add_cabinet_args (and optionally add_topology_args).
	
	Get information about the dimensions of the cabinets in the system from the
	supplied arguments.
	
	Returns
	-------
	(:py:class:`spinner.cabinet.Cabinet`, num_frames)
		num_frames is the number of frames (per cabinet) to actually fill with
		boards.
	"""
	kwargs = {
		kw: tuple(getattr(args, kw))
		    if type(getattr(args, kw)) is list
		    else getattr(args, kw)
		for kw in [
			"board_dimensions",
			"board_wire_offset_south_west",
			"board_wire_offset_north_east",
			"board_wire_offset_east",
			"board_wire_offset_west",
			"board_wire_offset_north",
			"board_wire_offset_south",
			"inter_board_spacing",
			"boards_per_frame",
			"frame_dimensions",
			"frame_board_offset",
			"inter_frame_spacing",
			"frames_per_cabinet",
			"cabinet_dimensions",
			"cabinet_frame_offset",
			"inter_cabinet_spacing",
		]
	}
	
	# Work out number of boards to allow checking of num_cabinets and num_frames
	# (only possible if topology args are present)
	if hasattr(args, "num_boards") and args.num_boards is not None:
		num_boards = args.num_boards
	elif hasattr(args, "triads") and args.triads is not None:
		num_boards = 3 * args.triads[0] * args.triads[1]
	else:
		num_boards = None  # unknown!
	
	if args.num_cabinets is None and args.num_frames is None:
		# Try to pick an sensible default value if number of boards is known,
		# otherwise default to a single cabinet system.
		if num_boards is not None:
			num_cabinets, num_frames = min_num_cabinets(num_boards,
			                                            args.frames_per_cabinet,
			                                            args.boards_per_frame)
		else:
			num_cabinets = 1
			num_frames = args.frames_per_cabinet
	else:
		# Default to 1 cabinet
		if args.num_cabinets is None:
			num_cabinets = 1
		else:
			num_cabinets = args.num_cabinets
		
		# Default to the full number of frames
		if args.num_frames is None:
			num_frames = args.frames_per_cabinet
		else:
			num_frames = args.num_frames
			if num_frames > args.frames_per_cabinet:
				parser.error("more frames specified than fit in a cabinet")
			if num_cabinets > 1 and num_frames != args.frames_per_cabinet:
				parser.error("--num-frames must equal --frames-per-cabinet "
				             "when there is more than one cabinet")
	
	# Check that the number of cabinets/frames is sufficient for the number of
	# boards present (if known)
	if (num_boards is not None and
	    num_cabinets * num_frames * args.boards_per_frame < num_boards):
		parser.error("not enough cabinets/frames available for {} "
		             "boards".format(num_boards))
	
	kwargs["num_cabinets"] = num_cabinets
	
	try:
		return (Cabinet(**kwargs), num_frames)
	except ValueError as e:
		parser.error(e.args[0])


def add_histogram_args(parser):
	"""Add arguments for specifying the histogram of wire lengths (i.e.
	cabinets/frames) available for a given system."""
	histogram_group = parser.add_argument_group("wire length histogram options")
	histogram_mut_group = histogram_group.add_mutually_exclusive_group()
	histogram_mut_group.add_argument("--histogram-bins", "-H", type=int, metavar="N",
	                                 default=5,
	                                 help="number of bins to pack wire lengths into "
	                                      "in the histogram of wire lengths (default: "
	                                      "%(default)s)")


def get_histogram_from_args(parser, args):
	"""To be used with add_histogram_args.
	
	Check that the supplied arguments are valid and returns the number of bins
	requested.
	
	Returns
	-------
	int
		The number of bins to use in the wire-length histogram.
	"""
	if args.histogram_bins < 1:
		parser.error("--histogram-bins must be at least 1")
	return args.histogram_bins


def add_wire_length_args(parser):
	"""Add arguments for specifying sets of wire lengths available for a given
	system."""
	wire_length_group = parser.add_argument_group("available wire lengths")
	wire_length_group.add_argument("--wire-length", "-l", type=float, metavar="L",
	                               action="append", nargs="+",
	                               help="specify one or more available wire "
	                                    "lengths in meters")
	wire_length_group.add_argument("--minimum-slack", type=float, metavar="H",
	                               default=0.05,
	                               help="the minimum slack to allow in a "
	                                    "wire connecting two boards in meters")


def get_wire_lengths_from_args(parser, args, mandatory=False):
	"""To be used with add_wire_length_args.
	
	Also used internally by get_histogram_from_args.
	
	Check that the supplied arguments are valid and then return the list of wire
	lengths supplied..
	
	Returns
	-------
	([float, ...], min_slack)
		Gives the wire lengths available (which may be empty) and the minimum slack
		requested.
	"""
	if args.wire_length is None:
		wire_lengths = []
	else:
		# Flatten list of lists
		args.wire_length = sum(args.wire_length, [])
		
		# Check signs and for duplicates
		seen = set()
		for wire_length in args.wire_length:
			if wire_length <= 0.0:
				parser.error("--wire-lengths must be positive and non-zero")
			if wire_length in seen:
				parser.error("wire length {} defined multiple times".format(wire_length))
			seen.add(wire_length)
		
		wire_lengths = sorted(args.wire_length)
	
	if args.minimum_slack < 0.0:
		parser.error("--minimum-slack must be positive")
	
	if mandatory and not wire_lengths:
		parser.error("At least one --wire-length argument must be provided.")
	
	return (wire_lengths, args.minimum_slack)


def add_image_args(parser):
	"""Add arguments for specifying output image filenames and dimensions."""
	image_group = parser.add_argument_group("image file parameters")
	image_group.add_argument("filename", type=str,
	                         help="filename to write the output to (.pdf or .png)")
	image_group.add_argument("width", type=float, nargs="?",
	                         help="width of the image in mm for PDF and pixels for "
	                              "PNG (defaults to 280 mm if PDF and 1000 px for PNG)")
	image_group.add_argument("height", type=float, nargs="?",
	                         help="height of the image in mm for PDF and pixels for "
	                              "PNG (if only width is given, output will be at "
	                              "most width wide and width tall)")
	


def get_image_from_args(parser, args, aspect_ratio=1.0):
	"""To be used with add_image_args.
	
	Check that the supplied arguments are valid and then return the filename, type
	and dimensions of the image to be created.
	
	Parameters
	----------
	aspect_ratio : float
		If the user does not fully specify the image size, what aspect ratio
		(width/height) should the image defined be?
	
	Returns
	-------
	output_filename, file_type, image_width, image_height
	"""
	output_filename = args.filename
	
	# Detect file-type from file-name
	if args.filename.lower().endswith(".png"):
		file_type = "png"
	elif args.filename.lower().endswith(".pdf"):
		file_type = "pdf"
	else:
		parser.error("filename must end in .png or .pdf")
	
	# Determine the size of the output image
	if args.width is None and args.height is None:
		# Width and height not specified, use default sizes
		if file_type == "png":
			args.width = 1000
		else:  # if file_type == "png":
			args.width = 280.0
	
	if args.width is not None and args.height is None:
		# Height is not specified, make the longest side as long as width and make
		# the whole image the same aspect ratio as the part of the system we're
		# focusing on
		if aspect_ratio < 1.0:
			# Tall image
			args.height = args.width
			args.width = args.height * aspect_ratio
		else:
			# Wide image
			args.height = args.width / aspect_ratio
	
	# Clamp to integers if PNG
	if file_type == "png":
		args.width = int(args.width)
		args.height = int(args.height)
	
	image_width = args.width
	image_height = args.height
	
	# Image dimensions must be non-zero and positive
	if image_width <= 0 or image_height <= 0:
		parser.error("image dimensions must be greater than 0")
	
	return (output_filename, file_type, image_width, image_height)


def add_bmp_args(parser):
	"""Add arguments for specifying a set of BMPs to connect to."""
	bmp_group = parser.add_argument_group("SpiNNaker BMP connection details")
	bmp_group.add_argument("--bmp", type=str, nargs=3, action="append",
	                       metavar=("CABINET", "FRAME", "HOSTNAME"),
	                       help="specify the hostname of a BMP to use to "
	                            "communicate with SpiNNaker boards in the given "
	                            "frame")


def get_bmps_from_args(parser, args, num_cabinets, num_frames):
	"""To be used with add_bmp_args.
	
	Check that the supplied arguments are valid and then return either a
	dictionary mapping (cabinet, frame) tuples to hostnames or a (hostname, port)
	tuple indicating the .
	
	Either no arguments must be supplied or a proxy client or exactly one per
	frame.
	"""
	# Special case when no arguments supplied.
	if args.bmp is None:
		return {}
	else:
		bmp_hostnames = {}
		
		for cabinet, frame, hostname in args.bmp:
			# Check types
			if not cabinet.isdigit():
				parser.error("--bmp cabinet number must be a number")
			cabinet = int(cabinet)
			
			if not frame.isdigit():
				parser.error("--bmp frame number must be a number")
			frame = int(frame)
			
			# Check for duplicates
			if (cabinet, frame) in bmp_hostnames:
				parser.error("bmp hostname for cabinet {}, frame {} "
				             "specified multiple times".format(cabinet, frame))
			if hostname in itervalues(bmp_hostnames):
				parser.error("bmp hostname '{}' given for more than one frame".format(
					hostname))
			
			bmp_hostnames[(cabinet, frame)] = hostname
		
		# Check every frame is included
		missing = set((c, f)
		              for c in range(num_cabinets)
		              for f in range(num_frames)) - set(bmp_hostnames)
		extra = set(bmp_hostnames) - set((c, f)
		                                 for c in range(num_cabinets)
		                                 for f in range(num_frames))
		
		if missing:
			parser.error("BMP hostname missing for {}".format(  # pragma: no branch
				", ".join("C:{} F:{}".format(c, f) for c, f in missing)))
		elif extra:
			parser.error(  # pragma: no branch
				"unexpected BMP for {} which are not part of the system".format(
					", ".join("C:{} F:{}".format(c, f) for c, f in extra)))
		
		return bmp_hostnames


def add_proxy_args(parser):
	"""Add arguments for specifying a proxy server connect to."""
	proxy_group = parser.add_argument_group("SpiNNaker proxy connection details")
	proxy_group.add_argument("--proxy", type=str, metavar="HOSTNAME",
	                         help="specify the hostname of a "
	                              "spinner-proxy-server instance to use "
	                              "to communicate with the SpiNNaker system")
	proxy_group.add_argument("--proxy-port", type=int, default=DEFAULT_PORT,
	                         metavar="PORT",
	                         help="specify the port to connect to "
	                              "spinner-proxy-server with (default: "
	                              "%(default)d)")


def get_proxy_from_args(parser, args):
	"""To be used with add_proxy_args.
	
	Returns a (hostname, port) or None if no proxy server details provided.
	
	This call checks that no conflicting --bmp arguments are present.
	"""
	# Make sure only one of the --bmp or --proxy arguments is given
	if hasattr(args, "bmp"):
		if args.bmp is not None and args.proxy is not None:
			parser.error("--bmp and --proxy are mutually exclusive")
	
	# If proxy specified, return one
	if args.proxy is not None:
		return (args.proxy, args.proxy_port)
	else:
		return None


def add_subset_args(parser):
	"""Add arguments for specifying a subset of wires to connect."""
	subset_group = parser.add_argument_group(
		"wire subset selection",
		description="""
			These arguments allow the specificiation of subsets of wires
			to install, for example, selecting only particular wires
			within a particular cabinet or frame. If no subsets are
			specified, all wires will be included, otherwise the union
			of all specified subsets are included. Use '1.2.*' to select all wires
			between boards in cabinet 1, frame 2. Use '1.*.*' to select all wires
			between boards in cabinet 1. Use '1-2.*.*' to select all
			wires which cross between cabinets 1 and 2.
		"""
	)
	subset_group.add_argument("--subset", nargs="+", type=str, metavar="SUBSET",
	                    help="specify the subset of wires to include")


def get_subset_from_args(parser, args):
	"""To be used with add_subset_args.
	
	Check that the supplied arguments are valid and then return a function with
	the prototype fn(wire) -> bool where wire is a tuple ((c, f, b, d), (c, f, b,
	d), length).
	"""
	if args.subset is None:
		# Special case when no arguments supplied: accept everything
		return (lambda wire: True)
	else:
		rules = []
		for subset in args.subset:
			# Split into cabinet, frame and board rules
			parts = re.split(r"\s*[,.]\s*", subset)
			if len(parts) != 3:
				parser.error(
					"--subset arguments must be of the form '???.???.???'"
					", not {}".format(subset))
			
			rule_fragments = []
			for i, part in enumerate(parts):
				# For each part, work out the rule
				match = re.match(r"^("
				                   r"(?P<single>[0-9]+)|"  # Specific c/f/b
				                   r"((?P<from>[0-9]+)-(?P<to>[0-9]+))|"  # Between two c/f/b
				                   r"(?P<wildcard>[*])"  # Wildcard
				                 r")$", part)
				if not match:
					parser.error(
						"--subset components must be either 'N', 'N-M' or '*', "
						"not {}".format(part))
				elif match.group("single"):
					value = int(match.group("single"))
					rule_fragments.append(
						lambda wire, i=i, value=value: (wire[0][i] == value and
					                                  wire[1][i] == value))
				elif match.group("from") and match.group("to"):
					frm = int(match.group("from"))
					to = int(match.group("to"))
					rule_fragments.append(
						lambda wire, i=i, frm=frm, to=to: (wire[0][i] == frm and
					                                     wire[1][i] == to)
					                                    or (wire[1][i] == frm and
					                                        wire[0][i] == to))
				elif match.group("wildcard"):
					# Don't append any rules: the wildcard always matches
					pass
				else:  # pragma: no cover
					assert False
			
			rules.append(lambda wire, rule_fragments=rule_fragments:
			             all(rf(wire) for rf in rule_fragments))
		
		return (lambda wire: any(r(wire) for r in rules))




if __name__=="__main__":  # pragma: no cover
	# This file when run as a script acts as a quick proof-of-concept of all
	# argument parsing capabilities
	import argparse
	parser = argparse.ArgumentParser()
	add_image_args(parser)
	add_topology_args(parser)
	add_cabinet_args(parser)
	add_histogram_args(parser)
	add_bmp_args(parser)
	add_subset_args(parser)
	
	args = parser.parse_args()
	print(get_image_from_args(parser, args, 0.5))
	print(get_topology_from_args(parser, args))
	print(get_cabinets_from_args(parser, args))
	print(get_histogram_from_args(parser, args))
	print(get_bmps_from_args(parser, args, 2, 2))
	print(get_subset_from_args(parser, args))



#!/usr/bin/env python

"""
A representation of Boards linked by wires in a SpiNNaker system along with
utilities for creating systems of them and iterating over them.
"""

from six import iteritems

from spinner.topology import Direction, \
	threeboards, wrap_around, add_direction

from spinner import coordinates


class Board(object):
	"""
	Represents a SpiNNaker board in a complete system.
	
	A board is an entity with links to its six neighbouring boards.
	"""
	
	# Counter used to label boards
	NEXT_BOARD_ID = 0
	
	def __init__(self):
		
		# References to other boards in the system which lie at the end of a wire
		# connected to a particular port.
		self.connection = {
			Direction.north      : None,
			Direction.north_east : None,
			Direction.east       : None,
			Direction.south      : None,
			Direction.south_west : None,
			Direction.west       : None,
		}
		
		# Set the board's ID
		self.id = Board.NEXT_BOARD_ID
		Board.NEXT_BOARD_ID += 1
	
	
	def connect_wire(self, other, direction):
		"""
		Connect a wire between this board and another for the given direction.
		"""
		# Ensure it isn't already connected
		assert(self.follow_wire(direction) is None)
		assert(other.follow_wire(direction.opposite) is None)
		
		self.connection[direction] = other
		other.connection[direction.opposite] = self
	
	
	def follow_wire(self, direction):
		"""
		Follow the wire going in the given direction from this board.
		"""
		return self.connection[direction]
	
	
	def follow_packet(self, in_wire_side, packet_direction):
		"""
		Follow the path of a packet which entered in to the board via the wire
		in_wire_side following packet_direction through the chips in the board.
		Returns a tuple (next_in_wire_side, next_board).
		
		We only need to know the side on which the incoming link is on (not the
		exact chip) because for any incoming side there is a fixed outgoing side
		when travelling in a fixed direction.
		"""
		
		# Mapping of {(in_wire_side, packet_direction) : out_wire_side,...}
		out_sides = {
			(Direction.south_west, Direction.east)       : Direction.east,
			(Direction.west,       Direction.east)       : Direction.north_east,
			
			(Direction.south_west, Direction.north_east) : Direction.north,
			(Direction.south,      Direction.north_east) : Direction.north_east,
			
			(Direction.south,      Direction.north)      : Direction.west,
			(Direction.east,       Direction.north)      : Direction.north,
		}
		# Opposite cases are simply inverted versions of the above...
		for (iws, pd), ows in iteritems(out_sides.copy()):
			out_sides[( iws.opposite
			          , pd.opposite
			          )] = ows.opposite
		
		out_wire_side = out_sides[(in_wire_side, packet_direction)]
		
		return (out_wire_side.opposite, self.follow_wire(out_wire_side))
		
	
	def __repr__(self):
		return "<Board id={}>".format(self.id)



def create_torus(width = 1, height = None):
	"""
	Returns a mapping of boards containing width * height threeboards connected in
	a torus with corresponding hexagonal coordinates. If height is not specified,
	height = width.
	"""
	
	height = width if height is None else height
	
	boards = {}
	
	# Create the boards
	for coord in threeboards(width, height):
		boards[coordinates.Hexagonal(*coord)] = Board()
	
	# Link the boards together
	for coord in boards:
		for direction in [ Direction.east
		                 , Direction.north_east
		                 , Direction.north
		                 ]:
			# Get the coordinate of the neighbour in each direction
			n_coord = wrap_around(
			            add_direction(list(coord)+[0], direction), (width, height))
			
			# Connect the boards together
			boards[coord].connect_wire(boards[n_coord], direction)
	
	return [(b, c) for (c, b) in iteritems(boards)]



#!/usr/bin/env python

"""
A tool which enumerates all connections to be made in a machine.
"""

import sys

import argparse

from spinner.utils import folded_torus

from spinner import transforms

from spinner.scripts import arguments

from spinner.plan import generate_wiring_plan, flatten_wiring_plan


def main(args=None):
	parser = argparse.ArgumentParser(
		description="Textually enumerate every connection required in a machine.")
	
	parser.add_argument("--sort-by", "-s",
	                    choices=["installation-order", "board", "wire-length"],
	                    default="board",
	                    help="Specifies the order the connections should be "
	                         "listed in the file: installation-order sorts in "
	                         "the most sensible order for installation, "
	                         "board lists wires on a board-by-board basis, "
	                         "wire-length lists in order of wire length.")
	
	arguments.add_version_args(parser)
	
	arguments.add_topology_args(parser)
	arguments.add_cabinet_args(parser)
	arguments.add_wire_length_args(parser)
	
	# Process command-line arguments
	args = parser.parse_args(args)
	(w, h), transformation, uncrinkle_direction, folds =\
		arguments.get_topology_from_args(parser, args)
	
	cabinet, num_frames = arguments.get_cabinets_from_args(parser, args)
	
	wire_lengths, min_slack = arguments.get_wire_lengths_from_args(
		parser, args, mandatory=True)
	
	# Generate folded system
	hex_boards, folded_boards = folded_torus(w, h,
	                                         transformation,
	                                         uncrinkle_direction,
	                                         folds)
	
	# Divide into cabinets
	cabinetised_boards = transforms.cabinetise(folded_boards,
	                                           cabinet.num_cabinets,
	                                           num_frames,
	                                           cabinet.boards_per_frame)
	cabinetised_boards = transforms.remove_gaps(cabinetised_boards)
	physical_boards = transforms.cabinet_to_physical(cabinetised_boards, cabinet)
	
	# Generate wiring plan
	wires_between_boards, wires_between_frames, wires_between_cabinets =\
		generate_wiring_plan(cabinetised_boards, physical_boards,
		                     cabinet.board_wire_offset, wire_lengths, min_slack)
	flat_wiring_plan = flatten_wiring_plan(wires_between_boards,
	                                       wires_between_frames,
	                                       wires_between_cabinets,
	                                       cabinet.board_wire_offset)
	
	# Convert wiring plan into cabinet coordinates
	b2c = dict(cabinetised_boards)
	wires = []
	for ((src_board, src_direction), (dst_board, dst_direction), wire_length) \
	    in flat_wiring_plan:
		
		sc, sf, sb = b2c[src_board]
		dc, df, db = b2c[dst_board]
		wires.append(((sc, sf, sb, src_direction),
		              (dc, df, db, dst_direction),
		              wire_length, src_board, dst_board))
	
	b2p = dict(physical_boards)
	
	# Order as requested on the command-line
	if args.sort_by == "board":
		wires = sorted(wires)
	elif args.sort_by == "wire-length":
		wires = sorted(wires, key=(lambda w: (w[2], w[:2])))
	elif args.sort_by == "installation-order":  # pragma: no branch
		pass  # List is initially in assembly order
	
	print("C  F  B  Socket      C  F  B  Socket      Length")
	print("-- -- -- ----------  -- -- -- ----------  ------")
	for ((sc, sf, sb, src_direction), (dc, df, db, dst_direction), wire_length, src_board, dst_board) in wires:
		print("{:2d} {:2d} {:2d} {:10s}  {:2d} {:2d} {:2d} {:10s}  {:0.2f}".format(
			sc, sf, sb, src_direction.name.replace("_", " "),
			dc, df, db, dst_direction.name.replace("_", " "),
			wire_length))
	
	return 0


if __name__=="__main__":  # pragma: no cover
	import sys
	sys.exit(main())

